#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on February 12, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import re
from contextlib import contextmanager

from traits.api import (
    Dict, Enum, Event, HasStrictTraits, Instance, Int, List, String, WeakRef)

from karabo.common.api import InstanceStatus
from karabogui.itemtypes import NavigationItemTypes

DOMAIN_LEVEL = 0
TYPE_LEVEL = 1
MEMBER_LEVEL = 2


class DeviceTreeNode(HasStrictTraits):
    """A node in the DeviceTree
    """
    node_id = String
    server_id = String
    status = Enum(*InstanceStatus)
    capabilities = Int
    interfaces = Int
    attributes = Dict
    level = Int(-1)

    parent = WeakRef('DeviceTreeNode')
    children = List(Instance('DeviceTreeNode'))

    def child(self, node_id):
        for child in self.children:
            if child.node_id == node_id:
                return child
        return None

    def info(self):
        level = self.level
        if level == DOMAIN_LEVEL:
            return {'type': NavigationItemTypes.UNDEFINED,
                    'domainId': self.node_id}
        elif level == TYPE_LEVEL:
            return {'type': NavigationItemTypes.UNDEFINED,
                    'typeId': self.node_id}
        elif level == MEMBER_LEVEL:
            return {'type': NavigationItemTypes.DEVICE,
                    'deviceId': self.node_id,
                    'serverId': self.server_id,
                    'capabilities': self.capabilities,
                    'archive': self.attributes.get('archive', False)}

    def row(self):
        if self.parent is None:
            return 0
        return self.parent.children.index(self)


class DeviceSystemTree(HasStrictTraits):
    # Base node for the whole tree
    root = Instance(DeviceTreeNode, args=())

    # An event which is triggered the device statuses where updated
    status_update = Event

    # A context manager to enter when manipulating the tree
    update_context = Instance(object)

    # device/server node lookup dicts
    _device_nodes = Dict

    def clear_all(self):
        """Removes all data from the model.
        """
        self._device_nodes.clear()

        with self.update_context.reset_context():
            self.root.children = []

    def get_instance_node(self, instance_id):
        """Retrieve a tree node for a device instance"""
        return self._device_nodes.get(instance_id, None)

    def remove_device(self, instance_id):
        """Remove the entry for a device from the tree
        """
        node = self._device_nodes.pop(instance_id, None)
        if node is None:
            return False

        type_node = node.parent
        with self.update_context.removal_context(node):
            type_node.children.remove(node)

        if not type_node.children:
            domain_node = type_node.parent
            with self.update_context.removal_context(type_node):
                domain_node.children.remove(type_node)

            if not domain_node.children:
                self._remove_root_children(domain_node)

        return True

    def initialize(self, system_hash):
        with self.update_context.reset_context():
            self._handle_device_data('device', system_hash, append=False)

    def update(self, system_hash):
        with self.update_context.layout_context():
            self._handle_device_data('device', system_hash)

    def instance_update(self, system_hash):
        if "device" not in system_hash:
            return

        device_status = set()
        for device_id, _, attrs in system_hash['device'].iterall():
            device_node = self._device_nodes.get(device_id)
            if device_node is None:
                continue

            status = InstanceStatus(attrs['status'])
            device_node.status = status
            device_node.attributes = attrs
            device_status.add(device_id)

        if device_status:
            self.status_update = device_status

    def visit(self, visitor):
        """Walk each node of the device tree until visitor returns true

        Call the `visitor` function on each item, if the function returns
        a true-ish value, the loop will be stopped.
        """

        def _iter_tree_node(node):
            yield node
            for child in node.children:
                yield from _iter_tree_node(child)

        for t_node in _iter_tree_node(self.root):
            if visitor(t_node):
                break

    # ------------------------------------------------------------------

    def _update_context_default(self):
        """Traits default initializer for the `update_context` trait.
        """

        class Dummy:
            @contextmanager
            def reset_context(self):
                yield

            @contextmanager
            def insertion_context(self, parent_node, first, last):
                yield

            @contextmanager
            def removal_context(self, tree_node):
                yield

            @contextmanager
            def layout_context(self):
                yield

            @contextmanager
            def insert_root_context(self, first, last):
                yield

            @contextmanager
            def remove_root_context(self, first, last):
                yield

            @contextmanager
            def removal_children_context(self, tree_node):
                yield

        return Dummy()

    def _append_child_node(self, parent_node, child_node):
        # First and last indices, INCLUSIVE
        first = len(parent_node.children)
        last = first

        with self.update_context.insertion_context(parent_node, first, last):
            parent_node.children.append(child_node)

    def _set_child_node(self, parent_node, child_node):
        parent_node.children.append(child_node)

    def _set_root_children(self, node, append):
        """Set the a domain node in the root element"""
        if append:
            first = len(self.root.children)
            last = first
            with self.update_context.insert_root_context(first, last):
                self.root.children.append(node)
        else:
            self.root.children.append(node)

    def _remove_root_children(self, node):
        """Set a node in the root element"""
        first = node.row()
        last = first
        with self.update_context.remove_root_context(first, last):
            self.root.children.remove(node)

    # ------------------------------------------------------------------

    def _handle_device_data(self, device_type, system_hash, append=True):
        assert device_type in ('device', 'macro')
        handler = self._append_child_node if append else self._set_child_node

        if device_type not in system_hash:
            return

        for karabo_name, _, attrs in system_hash[device_type].iterall():
            if len(attrs) == 0:
                continue
            # If we don't follow karabo naming convention, we are out!
            device = karabo_name.split('/')
            if not len(device) == 3:
                continue

            domain, dev_type, member = device
            capabilities = attrs['capabilities']
            server_id = attrs['serverId']
            status = InstanceStatus(attrs['status'])
            # Interfaces are optional ...
            interfaces = attrs.get('interfaces', 0)

            domain_node = self.root.child(domain)
            if domain_node is None:
                domain_node = DeviceTreeNode(node_id=domain,
                                             parent=self.root,
                                             level=DOMAIN_LEVEL)
                self._set_root_children(domain_node, append)

            type_node = domain_node.child(dev_type)
            if type_node is None:
                type_node = DeviceTreeNode(node_id=dev_type,
                                           parent=domain_node,
                                           level=TYPE_LEVEL)
                handler(domain_node, type_node)

            device_node = type_node.child(karabo_name)
            if device_node is None:
                device_node = DeviceTreeNode(node_id=karabo_name,
                                             parent=type_node,
                                             server_id=server_id,
                                             status=status,
                                             attributes=attrs,
                                             capabilities=capabilities,
                                             interfaces=interfaces,
                                             level=MEMBER_LEVEL)
                handler(type_node, device_node)

            self._device_nodes[karabo_name] = device_node

    def find(self, node_id, case_sensitive=True,
             use_reg_ex=True, full_match=False):
        """Find all nodes with the given `node_id` and return them in a list

        :param node_id: The actual string we are looking for in the tree
        :param case_sensitive: States whether the given string should match
                               case or not
        :param full_match: States whether the given string matches fully
        :param use_reg_ex: Defines the given string as a regular expression
        :return list of found nodes (which could be empty)
        """
        found_nodes = []
        pattern = node_id if use_reg_ex else f".*{re.escape(node_id)}"
        flags = 0 if case_sensitive else re.IGNORECASE
        regex = re.compile(pattern, flags=flags)

        def visitor(node):
            parent_node = node.parent
            parent_check = parent_node is None
            if parent_check:
                # Do not look for nodes which are its parent
                return

            matcher = regex.fullmatch if full_match else regex.match
            if matcher(node.node_id) is not None:
                found_nodes.append(node)

        self.visit(visitor)

        return found_nodes



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest
from traits.api import TraitError

from karabogui.binding.api import (
    BoolBinding, FloatBinding, Int8Binding, Int16Binding, Int32Binding,
    Int64Binding, Uint8Binding, Uint16Binding, Uint32Binding, Uint64Binding)


def test_validate_integers():
    """Test the coerce of our numpy integer range"""

    bindings = {
        Int64Binding: np.int64,
        Int32Binding: np.int32,
        Int16Binding: np.int16,
        Int8Binding: np.int8,
        Uint64Binding: np.uint64,
        Uint32Binding: np.uint32,
        Uint16Binding: np.uint16,
        Uint8Binding: np.uint8,
    }

    for property_binding, dtype in bindings.items():
        binding = property_binding()
        value = binding.validate_trait("value", dtype(23))
        assert value == 23
        assert isinstance(value, dtype)
        value = binding.validate_trait("value", "23")
        assert value == 23
        assert isinstance(value, dtype)
        value = binding.validate_trait("value", 2)
        assert value == 2
        assert isinstance(value, dtype)
        # Put a goofy value and check raise
        with pytest.raises(TraitError):
            binding.validate_trait("value", [])

    # Check basic limits for unsigned ints again ...
    bindings = {
        Uint64Binding: np.uint64,
        Uint32Binding: np.uint32,
        Uint16Binding: np.uint16,
        Uint8Binding: np.uint8,
    }
    for property_binding in bindings.keys():
        binding = property_binding()
        negative_value = -1
        with pytest.raises(TraitError):
            binding.validate_trait("value", negative_value)

        negative_overflow = -2e100
        with pytest.raises(TraitError):
            binding.validate_trait("value", str(negative_overflow))
        with pytest.raises(TraitError):
            binding.validate_trait("value", negative_overflow)

        positive_overflow = 2e100
        with pytest.raises(TraitError):
            binding.validate_trait("value", str(positive_overflow))
        with pytest.raises(TraitError):
            binding.validate_trait("value", positive_overflow)


def test_validate_boolean():
    """Test if the coerce of the bool binding"""

    binding = BoolBinding()
    value = binding.validate_trait("value", 23)
    assert value is True
    assert isinstance(value, bool)
    value = binding.validate_trait("value", "23")
    assert value is True
    assert isinstance(value, bool)
    value = binding.validate_trait("value", False)
    assert value is False
    assert isinstance(value, bool)
    value = binding.validate_trait("value", True)
    assert value is True
    assert isinstance(value, bool)
    value = binding.validate_trait("value", 0)
    assert value is False
    assert isinstance(value, bool)
    value = binding.validate_trait("value", 1)
    assert value is True
    assert isinstance(value, bool)
    with pytest.raises(Exception):
        binding.validate_trait("value", "None")
    with pytest.raises(Exception):
        binding.validate_trait("value", None)


def test_validate_float():
    """Test if the coerce of the float binding"""

    binding = FloatBinding()
    value = binding.validate_trait("value", 23)
    assert value == 23.0
    assert isinstance(value, float)
    value = binding.validate_trait("value", "23")
    assert value == 23
    assert isinstance(value, float)
    value = binding.validate_trait("value", np.float32(4))
    assert value == 4.0
    assert isinstance(value, np.float32)
    value = binding.validate_trait("value", np.float64(4))
    assert value == 4.0
    assert isinstance(value, np.float64)


def test_numpy_binding_types():
    """Test if our numpy value type is preserved"""
    int64Property = Int64Binding()
    int64Property.value = np.int64(23)
    assert int64Property.value == 23
    assert isinstance(int64Property.value, np.int64)
    int64Property.value = 64
    assert int64Property.value == 64
    assert isinstance(int64Property.value, np.int64)
    int64Property.value = 23
    assert int64Property.value == 23
    assert isinstance(int64Property.value, np.int64)

    int32Property = Int32Binding()
    int32Property.value = np.int32(23)
    assert int32Property.value == 23
    assert isinstance(int32Property.value, np.int32)
    int32Property.value = 64
    assert int32Property.value == 64
    assert isinstance(int32Property.value, np.int32)
    int32Property.value = np.int16(16)
    assert int32Property.value == 16
    assert isinstance(int32Property.value, np.int32)

    int16Property = Int16Binding()
    int16Property.value = np.int16(23)
    assert int16Property.value == 23
    assert isinstance(int16Property.value, np.int16)
    int16Property.value = 64
    assert int16Property.value == 64
    assert isinstance(int16Property.value, np.int16)
    int16Property.value = np.int8(16)
    assert int16Property.value == 16
    assert isinstance(int16Property.value, np.int16)

    int8Property = Int8Binding()
    int8Property.value = np.int8(23)
    assert int8Property.value == 23
    assert isinstance(int8Property.value, np.int8)
    int8Property.value = 64
    assert int8Property.value == 64
    assert isinstance(int8Property.value, np.int8)
    int8Property.value = np.int16(16)
    assert int8Property.value == 16
    assert isinstance(int8Property.value, np.int8)

    uint64Property = Uint64Binding()
    uint64Property.value = np.uint64(23)
    assert uint64Property.value == 23
    assert isinstance(uint64Property.value, np.uint64)
    uint64Property.value = 64
    assert uint64Property.value == 64
    assert isinstance(uint64Property.value, np.uint64)
    uint64Property.value = np.int16(16)
    assert uint64Property.value == 16
    assert isinstance(uint64Property.value, np.uint64)

    uint32Property = Uint32Binding()
    uint32Property.value = np.uint32(23)
    assert uint32Property.value == 23
    assert isinstance(uint32Property.value, np.uint32)
    uint32Property.value = 64
    assert uint32Property.value == 64
    assert isinstance(uint32Property.value, np.uint32)
    uint32Property.value = np.int16(16)
    assert uint32Property.value == 16
    assert isinstance(uint32Property.value, np.uint32)

    uint16Property = Uint16Binding()
    uint16Property.value = np.uint16(23)
    assert uint16Property.value == 23
    assert isinstance(uint16Property.value, np.uint16)
    uint16Property.value = 64
    assert uint16Property.value == 64
    assert isinstance(uint16Property.value, np.uint16)

    uint8Property = Uint8Binding()
    uint8Property.value = np.uint8(23)
    assert uint8Property.value == 23
    assert isinstance(uint8Property.value, np.uint8)
    uint8Property.value = 64
    assert uint8Property.value == 64
    assert isinstance(uint8Property.value, np.uint8)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from functools import partial

from traits.api import (
    Any, Callable, Constant, Dict, Enum, Event, HasStrictTraits, List)

from ..enums import ExportTool, MouseTool, ROITool
from .factory import export_factory, mouse_mode_factory, roi_factory


class BaseToolsetController(HasStrictTraits):
    default_tool = Any
    current_tool = Any

    tools = List
    factory = Callable

    buttons = Dict

    def _buttons_default(self):
        buttons = {}
        for tool in self.tools:
            button = self.factory(tool)
            if button is not None:
                button.clicked.connect(partial(self.select, tool=tool))
                buttons[tool] = button

        # If there's a default button, enable it
        default_button = buttons.get(self.default_tool)
        if default_button:
            default_button.setChecked(True)

        return buttons

    def select(self, tool):
        """Selecting a tool triggers the connected events."""
        self.current_tool = tool

    def add(self, tool):
        """Add an optional tool to the existing controller set"""
        assert tool not in self.tools
        button = self.factory(tool)
        if button is not None:
            self.buttons[tool] = button
            button.clicked.connect(partial(self.select, tool=tool))
            self.tools.append(tool)
        return button

    def check(self, tool):
        """Checking the tool only changes the check state of the button.
           We then update the current_tool quietly to not trigger any events"""
        if tool in self.buttons:
            self.buttons[tool].setChecked(True)
        self.trait_setq(current_tool=tool)

    def uncheck_all(self):
        for button in self.buttons.values():
            button.setChecked(False)
            # Uncheck default action if any:
            action = button.defaultAction()
            if action:
                action.setChecked(False)
        self.trait_setq(current_tool=self.default_tool)


class MouseModeToolset(BaseToolsetController):
    tools = List([MouseTool.Pointer, MouseTool.Zoom, MouseTool.Move])
    factory = Callable(mouse_mode_factory)

    current_tool = Enum(*MouseTool)
    default_tool = Constant(MouseTool.Pointer)

    def select(self, tool):
        """Sets the selected mouse mode and uncheck any previously tool"""

        # Uncheck current tool
        self.buttons[self.current_tool].setChecked(False)

        # If current tool is unchecked, select default tool
        if self.current_tool == tool and self.default_tool in self.buttons:
            tool = self.default_tool
            self.buttons[tool].setChecked(True)

        super().select(tool)


class ROIToolset(BaseToolsetController):
    tools = List([ROITool.Rect, ROITool.Crosshair])
    factory = Callable(roi_factory)

    current_tool = Enum(*ROITool)
    default_tool = ROITool.NoROI

    def select(self, tool):
        """The toolset has can have one or more buttons, with check states
           being exclusive. When a button is unchecked, the toolset returns
           ROI.NoROI. When selected, on the other hand, the other button
           (if existing), should be unchecked."""

        cross_button = self.buttons[ROITool.Crosshair]
        rect_button = self.buttons[ROITool.Rect]

        # Uncheck the other button
        if not cross_button.isChecked() and not rect_button.isChecked():
            tool = ROITool.NoROI
        elif tool in [ROITool.Rect, ROITool.DrawRect]:
            if rect_button.isChecked():
                cross_button.setChecked(False)
        elif tool in [ROITool.Crosshair, ROITool.DrawCrosshair]:
            if cross_button.isChecked():
                rect_button.setChecked(False)

        # Get the proper tool from the actions
        if tool in self.buttons:
            tool = self.buttons[tool].defaultAction().data()
        super().select(tool)

    def check(self, tool):
        """Checking the tool only changes the check state of the button.
           We then update the current_tool quietly to not trigger any events.

           Reimplementing since we want to change the default action"""

        # Uncheck all if no ROI is selected
        if tool == ROITool.NoROI:
            self.uncheck_all()
            return

        if tool in [ROITool.Rect, ROITool.DrawRect]:
            button = self.buttons[ROITool.Rect]
        elif tool in [ROITool.Crosshair, ROITool.DrawCrosshair]:
            button = self.buttons[ROITool.Crosshair]
        else:
            # Do nothing if no button is associated with the tool
            return

        button.setChecked(True)
        for action in button.menu().actions():
            if action.data() == tool:
                button.setDefaultAction(action)
            action.setChecked(action.data() == tool)

        self.trait_setq(current_tool=tool)


class ExportToolset(BaseToolsetController):

    tools = List([ExportTool.Image, ExportTool.Data])
    factory = Callable(export_factory)

    # Current tool is not persistent, as this is a fire-and-forget toolset
    current_tool = Event


TOOLSET_MAP = {
    ExportTool: ExportToolset,
    MouseTool: MouseModeToolset,
    ROITool: ROIToolset}


def get_toolset(toolset, tools=None):
    klass = TOOLSET_MAP.get(toolset)
    if klass is not None:
        if tools is not None:
            return klass(tools=tools)
        else:
            return klass()


def register_toolset(toolset, klass):
    if toolset not in TOOLSET_MAP:
        TOOLSET_MAP[toolset] = klass



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on March 6, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from urllib.parse import parse_qs

from qtpy.QtGui import QBrush, QColor

from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_STATE
from karabo.native import AccessMode, Hash, encodeBinary


def is_writable_binding(binding, binding_type=None):
    """Return if the `binding` is a writable string binding"""
    writable = binding.accessMode is AccessMode.RECONFIGURABLE
    if binding_type is None:
        return writable
    return isinstance(binding, binding_type) and writable


def is_state_display_type(binding):
    """Return if the display type belongs to a state element"""
    return binding.displayType == KARABO_SCHEMA_DISPLAY_TYPE_STATE


def string2list(value: str) -> list:
    """Convert a list of chars to a list of strings with delimiter ','.

    This function strips white spaces and returns an empty list if an
    empty string is passed.
    """
    return [v.strip() for v in value.split(",")] if value else []


def list2string(value: list) -> str:
    """Convert a list to a string representation with delimiter ','"""
    value = [] if value is None else value
    return ",".join(str(v) for v in value)


def create_brushes(display_string):
    """Parse a displayType string protocol for brushes

    e.g. a display string with text1=color1&text2=color2

    Returns a default brush and a dictionary of brushes
    """

    def create_brush(color):
        """Protective function to create a QBrush from a color"""
        try:
            return QBrush(QColor(color))
        except Exception:
            return None  # No brush!

    query_colors = display_string.split("|")
    if not len(query_colors) == 2:
        return None, {}

    parsed = parse_qs(query_colors[1], keep_blank_values=True)
    color_map = {name: color[0] for name, color in parsed.items()}
    default = color_map.pop("default", None)
    default_brush = create_brush(default) if default is not None else None

    brushes = {text: create_brush(color) for text, color
               in color_map.items()}

    return default_brush, brushes


def has_confirmation(display_string):
    attributes = get_button_attributes(display_string)
    return attributes.get("confirmation") == "1"


def get_button_attributes(display_string):
    """Retrieve the button attributes from a table display string"""
    splitted = display_string.split("|")
    if not len(splitted) == 2:
        return {}

    parsed = parse_qs(splitted[1], keep_blank_values=False)
    options = {name: opt[0] for name, opt in parsed.items()}

    return options


def parse_table_link(string):
    """Parse a table link for a table string button

    Valid schemes are for now `url` and `deviceScene`
    """
    splitted = string.split("|", 1)
    if not len(splitted) == 2:
        return None, {}

    action_type = splitted[0]
    if action_type == "deviceScene":
        parsed = parse_qs(splitted[1], keep_blank_values=False)
        options = {name: opt[0] for name, opt in parsed.items()}
    elif action_type == "url":
        options = splitted[1]
    else:
        options = {}
    return action_type, options


def quick_table_copy(hsh):
    """Return a quick deepcopy of a table hash

    This only works with elements that are supported in a table element
    """

    def _quick_copy(value):
        """Internal function to quickly provide a quick value copy"""
        try:
            # lists, tuples, unicode, arrays
            return value.copy()
        except AttributeError:
            # Simple types
            return value

    return Hash({k: _quick_copy(v) for k, v in hsh.items()})


def create_mime_data(**kwargs):
    """Return byte encoded mime data from kwargs"""
    return encodeBinary(Hash(kwargs))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtGui import QFont

from ..fonts import FONT_STYLE_HINTS, get_qfont

# This maps the font family to a Qt style hint to ease font replacement
SUGGESTED_STYLE_HINTS = {
    QFont.AnyStyle: "Sans Serif",
    QFont.Helvetica: "Helvetica",
    QFont.Times: "Times",
    QFont.Courier: "Courier"
}


def test_get_font_from_string():
    for style, family in FONT_STYLE_HINTS.items():
        # Check if font is loaded
        old_qfont = QFont(family)
        assert old_qfont.family() == family

        # Check if suggested font family is replaced by the application font
        # accordingly
        new_qfont = get_qfont(family)
        assert new_qfont.family() == FONT_STYLE_HINTS[style]



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on May 9, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from functools import partial
from inspect import signature
from io import StringIO

from qtpy.QtCore import QTimer, Slot
from traits.api import Undefined

from karabo.common.api import Capabilities, walk_traits_object
from karabo.common.project.api import read_macro
from karabo.common.scenemodel.api import SceneTargetWindow, read_scene
from karabo.native import Hash
from karabogui import messagebox
from karabogui.binding.api import DeviceProxy, extract_sparse_configurations
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.fonts import substitute_font
from karabogui.logger import get_logger
from karabogui.singletons.api import get_manager, get_network, get_topology
from karabogui.util import get_reason_parts

_WAIT_SECONDS = 5
_waiting_devices = {}


def call_device_slot(handler, instance_id, slot_name, **kwargs):
    """Call a device slot via the GUI server. This works with slots which
    take a single `Hash` as an argument and reply with a `Hash`.

    :param handler: Callable with the signature

                    - handler(success, reply)
                    - handler(success, reply, request)

                    where:
                        success - a bool indicating whether the call succeeded
                        reply - a Hash containing the slot reply
                        request - the initial request

                    This handler can be called when the GUI server connection
                    is lost. In that case, `success` is False, and `reply`
                    is the `reason` of failure (string).
                    In case the call fails or times out, `success` is also
                    False.
                    The request Hash can only include the `token` on failure

    :param instance_id: Device ID of the device whose slot will be called
    :param slot_name: Name of the slot which will be called
    :param **kwargs: The argument list of the slot being called. `token` is
                     allowed to be a key-word argument

    :returns token: A unique identifier for the call

    NOTE: You had better know what you are doing here. Misuse of this feature
    can seriously degrade performance of devices external to the GUI server.
    """
    # Verify that the handler takes the correct number of arguments
    sig = signature(handler)
    if len(sig.parameters) not in (2, 3):
        raise ValueError(
            'A slot callback handler must take either two or three arguments')

    # Prepare the parameters Hash
    params = Hash()
    for k, v in kwargs.items():
        params.set(k, v)

    # Call the slot and return the token
    return get_manager().callDeviceSlot(handler, instance_id,
                                        slot_name, params)


def send_property_changes(proxies):
    """Sends a reconfiguration request to the GUI server

    :param proxies: A sequence (list, tuple, set) of PropertyProxy instances
                    with their `edit_value` trait set. See the function
                    `extract_sparse_configurations` for more details.
    """
    configs = extract_sparse_configurations(proxies)
    network, manager, topology = get_network(), get_manager(), get_topology()
    for device_id, config in configs.items():
        properties = [proxy for proxy in proxies
                      if proxy.root_proxy.device_id == device_id]
        device_proxy = topology.get_device(device_id)
        manager.expect_properties(device_proxy, properties)
        network.onReconfigure(device_id, config)
        # and more information
        info = Hash("type", "reconfigure", "instanceId", device_id,
                    "configuration", config)
        network.onInfo(info)


def onConfigurationUpdate(proxy, handler=None, request=False, remove=True):
    """Execute on a config update of `proxy` and request if necessary

    :param proxy: The DeviceProxy or PropertyProxy
    :param handler: The handler to be executed on update
    :param request: Boolean to set if a config should be requested
    :param remove: Boolean to set if the handler should be removed after
    """
    if not isinstance(proxy, DeviceProxy):
        root_proxy = proxy.root_proxy
    else:
        root_proxy = proxy

    def config_handler():
        if remove:
            proxy.binding.on_trait_change(config_handler, "config_update",
                                          remove=True)
        handler()

    proxy.binding.on_trait_change(config_handler, "config_update")
    if request and root_proxy.online:
        device_id = root_proxy.device_id
        get_network().onGetDeviceConfiguration(device_id)

    return config_handler


def onSchemaUpdate(proxy, handler, request=False, remove=True):
    """Execute on a schema update of `proxy` and `request` if necessary

    :param proxy: The DeviceProxy or PropertyProxy
    :param handler: The handler to be executed on update
    :param request: Boolean to set if a config should be requested
    :param remove: Boolean to set if the handler should be removed after
    """
    if not isinstance(proxy, DeviceProxy):
        root_proxy = proxy.root_proxy
    else:
        root_proxy = proxy

    def schema_handler():
        if remove:
            root_proxy.on_trait_change(schema_handler, "schema_update",
                                       remove=True)
        handler()

    root_proxy.on_trait_change(schema_handler, "schema_update")
    if request and root_proxy.online:
        device_id = root_proxy.device_id
        get_network().onGetDeviceSchema(device_id)

    return schema_handler


def get_scene_from_server(device_id, scene_name, project=None,
                          target_window=SceneTargetWindow.Dialog,
                          slot_name='requestScene', **kwargs):
    """Get a scene from the a device

    :param device_id: The deviceId of the device
    :param scene_name: The scene name
    :param project: The project owner of the scene. Default is None.
    :param target_window: The target window option. Default is Dialog.
    :param slot_name: The slot to be called to retrieve the scene
    """

    def scene_handler(dev_id, name, project, target_window, success, reply,
                      request):
        """Callback handler for a request to a device"""

        if not success:
            reason, details = get_reason_parts(reply)
            msg = (f"Scene '{name}' from device '{dev_id}' was not retrieved."
                   "<br><br>The reason is:<br>"
                   f"<i>{reason}</i><br><br>")
            messagebox.show_warning(msg, title='Load Scene from Device Failed',
                                    details=details)
            return
        elif not reply.get('payload.success', False):
            reason = reply.get("payload.reason", "No reason specified")
            msg = (f"Retrieval of scene '{name}' from device '{dev_id}' "
                   f"reported an error: {reason}.")
            messagebox.show_warning(msg, title='Load Scene from Device Failed')
            return

        data = reply.get('payload.data', '')
        if not data:
            msg = 'Scene "{}" from device "{}" contains no data!'
            messagebox.show_warning(msg.format(name, dev_id),
                                    title='Load Scene from Device Failed')
            return

        with StringIO(data) as fp:
            scene = read_scene(fp)
            scene.assure_svg_data()
            scene.modified = True
            scene.simple_name = f'{dev_id}|{name}'
            scene.reset_uuid()

        # Add to the project AND open it
        event_type = KaraboEvent.ShowUnattachedSceneView
        window = SceneTargetWindow.MainWindow
        if target_window is not None:
            window = target_window
        if project is not None:
            event_type = KaraboEvent.ShowSceneView
            project.scenes.append(scene)

        def visitor(model):
            substitute_font(model)

        event_data = {'model': scene, 'target_window': window}
        position = request["args"].get("position")
        if position is not None:
            event_data["position"] = position

        walk_traits_object(scene, visitor_func=visitor)
        broadcast_event(event_type, event_data)

    handler = partial(scene_handler, device_id, scene_name, project,
                      target_window)
    return call_device_slot(handler, device_id, slot_name=slot_name,
                            name=scene_name, **kwargs)


def get_macro_from_server(device_id, macro_name, project):
    """Get a macro from a device

    :param device_id: The deviceId of the device
    :param macro_name: The macro name
    :param project: The project owner of the macro. Macros must have a project
    """

    def macro_handler(dev_id, name, project, success, reply):
        if not success or not reply.get('payload.success', False):
            msg = 'Macro "{}" from device "{}" was not retrieved!'
            messagebox.show_warning(msg.format(name, dev_id),
                                    title='Load Macro from Device Failed')
            return

        data = reply.get('payload.data', '')
        if not data:
            msg = 'Macro "{}" from device "{}" contains no data!'
            messagebox.show_warning(msg.format(name, dev_id),
                                    title='Load Macro from Device Failed')
            return

        with StringIO(data) as fp:
            macro = read_macro(fp)
            macro.initialized = macro.modified = True
            macro.simple_name = f'{dev_id}-{name}'
            macro.reset_uuid()

        # Macro's can only be added to project. Hence, add first to the project
        project.macros.append(macro)
        # and then open it
        broadcast_event(KaraboEvent.ShowMacroView, {'model': macro})

    handler = partial(macro_handler, device_id, macro_name, project)

    return call_device_slot(handler, device_id, slot_name='requestMacro',
                            name=macro_name)


def retrieve_default_scene(device_id):
    """Retrieve a default scene from device with `device_id`"""
    attrs = get_topology().get_attributes(f"device.{device_id}")
    if attrs is None:
        msg = f"The device <b>{device_id}</b> is not online."
        get_logger().info(msg)
        return

    capabilities = attrs.get("capabilities", 0)
    bit = Capabilities.PROVIDES_SCENES

    if (capabilities & bit) != bit:
        msg = f"The device <b>{device_id}</b> does not provide a scene."
        get_logger().info(msg)
        return

    proxy = get_topology().get_device(device_id)

    def config_update():
        proxy.on_trait_change(config_update, "config_update", remove=True)
        proxy.remove_monitor()
        get_proxy_scene(proxy, request=False)

    def get_proxy_scene(proxy, request=True):
        scenes = proxy["availableScenes"]
        if request and (scenes is None or scenes.value is Undefined):
            proxy.on_trait_change(config_update, "config_update")
            proxy.add_monitor()
        elif not len(scenes.value):
            get_logger().info(
                "The device <b>{}</b> does not specify a scene "
                "name!".format(device_id))
        else:
            scene_name = scenes.value[0]
            get_scene_from_server(device_id, scene_name)

    get_proxy_scene(proxy)


def onShutdown(device_proxy, handler, parent=None):
    """Shutdown a device_proxy and perform an action handler on device_proxy

    Note: The device must be online!
    """

    def _offline_handler(device_proxy, name, new):
        """Handle a device getting a new configuration
        """
        global _waiting_devices
        handler, timer = _waiting_devices.pop(device_proxy,
                                              (lambda: None, None))

        # Remove the trait handler
        device_proxy.on_trait_change(_offline_handler, "online", remove=True)

        # Clear the timer
        timer.stop()
        # Execute the handler
        handler()

    @Slot()
    def _timeout_handler(device_proxy):
        """Handle our wait timer expiring"""
        global _waiting_devices
        _waiting_devices.pop(device_proxy, None)

        # Remove the trait handler
        device_proxy.on_trait_change(_offline_handler, "online",
                                     remove=True)

        # Inform the user
        msg = f"The device didn't shutdown in {_WAIT_SECONDS} ... aborting."
        msg = msg.format(device_proxy.device_id, _WAIT_SECONDS)
        messagebox.show_warning(msg, parent=parent)

    if not device_proxy.online:
        msg = "The device is already offline ..."
        messagebox.show_error(msg, parent=parent)
        return

    global _waiting_devices
    if device_proxy in _waiting_devices:
        # Already waiting
        return

    timer = QTimer()
    timer.setSingleShot(True)
    timer.timeout.connect(partial(_timeout_handler, device_proxy))
    device_proxy.on_trait_change(_offline_handler, 'online')

    _waiting_devices[device_proxy] = (handler, timer)
    timer.start(_WAIT_SECONDS * 1000)

    instanceId = device_proxy.device_id
    # Shutdown the device
    get_network().onKillDevice(instanceId)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import DisplayFloatModel
from karabo.native import Configurable, Float, String
from karabogui.binding.api import (
    DeviceClassProxy, PropertyProxy, ProxyStatus, build_binding)
from karabogui.testing import set_proxy_value

from ..baselabel import BaseLabelController, FormatMixin


class Object(Configurable):
    string = String()
    decimals = Float(displayType="format|fmt=f&decimals=6")


def test_base_label_controller(gui_app, mocker):
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceClassProxy(binding=binding, server_id="Fake",
                              status=ProxyStatus.OFFLINE)
    string = PropertyProxy(root_proxy=device, path="string")
    decimals = PropertyProxy(root_proxy=device, path="decimals")

    model = DisplayFloatModel()
    controller = BaseLabelController(model=model, proxy=string)
    controller.create(None)
    assert controller.widget is not None
    set_proxy_value(string, "string", "hello")
    assert controller.internal_widget.text() == "hello"

    controller = BaseLabelController(model=model, proxy=decimals)
    controller.create(None)
    set_proxy_value(decimals, "decimals", 0.25)
    assert controller.internal_widget.text() == "0.25"

    action = controller.widget.actions()[0]
    assert action.text() == "Format field"
    dialog = "karabogui.controllers.baselabel.FormatLabelDialog"
    dialog = mocker.patch(dialog)
    dialog().font_size = 9
    dialog().font_weight = "bold"
    dialog().exec.return_value = QDialog.Accepted
    action.trigger()
    assert model.font_weight == "bold"
    assert model.font_size == 9


class BaseFloatController(FormatMixin, BaseLabelController):
    """A class to test the base format mixin"""


def test_format_mixin(gui_app, mocker):
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceClassProxy(binding=binding, server_id="Fake",
                              status=ProxyStatus.OFFLINE)
    decimals = PropertyProxy(root_proxy=device, path="decimals")

    model = DisplayFloatModel()
    controller = BaseFloatController(model=model, proxy=decimals)
    controller.create(None)
    assert controller.widget is not None
    set_proxy_value(decimals, "decimals", 10.25)
    # Default format is "{.8g}"
    assert controller.fmt == "{:.8g}"
    assert controller.internal_widget.text() == "10.25"

    # Format field, already tested
    action = controller.widget.actions()[0]
    assert action.text() == "Format field"

    # Formatting value dialog
    action = controller.widget.actions()[1]
    assert action.text() == "Format value"
    dialog = "karabogui.controllers.baselabel.FormatFmtDialog"
    dialog = mocker.patch(dialog)
    dialog().fmt = "f"
    dialog().decimals = "4"
    dialog().exec.return_value = QDialog.Accepted
    action.trigger()
    assert model.fmt == "f"
    assert model.decimals == "4"

    assert controller.fmt == "{:.4f}"
    assert controller.internal_widget.text() == "10.2500"

    # Check the drop of the float controller
    BaseFloatController.initialize_model(decimals, model)
    assert model.decimals == "6"
    assert model.fmt == "f"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import time

from qtpy.QtCore import QObject, QThreadPool, Slot

from karabogui.background import background

TASK_TIME = 0.1


def test_task_background(gui_app):
    """Test that we can run tasks in the background"""

    class Dialog(QObject):
        running = False
        result = None
        finished = False

        @Slot(object)
        def callback(self, future):
            self.result = future.result()
            self.finished = True

        def background_task(self):
            self.running = True
            time.sleep(TASK_TIME)
            return "Karabo"

        def long_task(self):
            self.running = True
            while self.running:
                time.sleep(TASK_TIME)
            return "Karabo"

    dialog = Dialog()
    background(dialog.background_task, callback=dialog.callback)
    QThreadPool.globalInstance().waitForDone(20)
    assert dialog.running
    assert not dialog.finished
    QThreadPool.globalInstance().waitForDone()
    gui_app.processEvents()
    assert dialog.result == "Karabo"
    assert dialog.finished

    dialog = Dialog()
    background(dialog.long_task, callback=dialog.callback)
    QThreadPool.globalInstance().waitForDone(20)
    assert dialog.running
    assert not dialog.finished
    dialog.running = False
    QThreadPool.globalInstance().waitForDone()
    gui_app.processEvents()
    assert dialog.result == "Karabo"


def test_run_background_functor(gui_app):
    """Test that we can cancel a runnable"""
    result = None

    @Slot(object)
    def callback(future):
        nonlocal result
        result = future.result()

    class Functor:
        running = False

        def __call__(self, *args, **kwargs):
            self.running = True
            while self.running:
                time.sleep(TASK_TIME)
            return "Karabo"

        def stop(self):
            self.running = False

    functor = Functor()
    background(functor, callback=callback)
    QThreadPool.globalInstance().waitForDone(20)
    gui_app.processEvents()
    assert functor.running
    QThreadPool.globalInstance().waitForDone(int(TASK_TIME * 1000))
    gui_app.processEvents()
    assert functor.running
    QThreadPool.globalInstance().waitForDone(int(TASK_TIME * 1000))
    gui_app.processEvents()
    assert functor.running
    functor.stop()
    QThreadPool.globalInstance().waitForDone()
    gui_app.processEvents()
    assert not functor.running
    assert result == "Karabo"


def test_run_background_exception(gui_app):
    """Test that we can redirect an exception to the main thread"""

    exception = None

    @Slot(object)
    def callback(future):
        nonlocal exception
        try:
            assert future.exception() is not None
            future.result()
        except RuntimeError as e:
            exception = e

    class Functor:
        running = False

        def __call__(self, *args, **kwargs):
            self.running = True
            time.sleep(TASK_TIME)
            raise RuntimeError("Device not installed")

    functor = Functor()
    background(functor, callback=callback)
    QThreadPool.globalInstance().waitForDone()
    gui_app.processEvents()
    assert exception is not None
    assert isinstance(exception, RuntimeError)
    assert str(exception) == "Device not installed"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.states import State
from karabogui.graph.common.const import ALL_STATES


def test_const_state_def():
    """Make sure that our own state mapping has all the values of Karabo"""
    state_defs = State.__members__.values()
    assert len(set(ALL_STATES)) == len(state_defs)

    for state in State:
        assert state in ALL_STATES



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import traceback

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QAction, QDialog, QFrame, QInputDialog, QLabel
from traits.api import Callable, Dict, Instance, Str, Tuple

from karabo.common.scenemodel.api import EvaluatorModel
from karabogui import messagebox
from karabogui.binding.api import (
    CharBinding, FloatBinding, IntBinding, StringBinding, get_binding_value)
from karabogui.const import WIDGET_MIN_HEIGHT
from karabogui.controllers.api import (
    BaseBindingController, add_unit_label, register_binding_controller)
from karabogui.dialogs.format_label import FormatLabelDialog
from karabogui.fonts import get_font_size_from_dpi
from karabogui.indicators import ALL_OK_COLOR
from karabogui.util import generateObjectName

BINDING_TYPES = (CharBinding, FloatBinding, StringBinding, IntBinding)


@register_binding_controller(ui_name='Evaluate Expression',
                             klassname='Evaluator',
                             binding_type=BINDING_TYPES)
class Evaluator(BaseBindingController):
    # The scene model class used by this controller
    model = Instance(EvaluatorModel, args=())
    # Security holes (evaluation of values)
    globals_ns = Dict
    function = Callable
    _bg_color = Tuple(ALL_OK_COLOR)
    _style_sheet = Str
    # Storage for the main widget
    _internal_widget = Instance(QLabel)

    def _globals_ns_default(self):
        ns = {}
        exec('from builtins import *', ns)
        exec('from numpy import *', ns)
        return ns

    def _function_default(self):
        return eval(f'lambda x: {self.model.expression}',
                    self.globals_ns)

    def create_widget(self, parent):
        self._internal_widget = QLabel(parent)
        self._internal_widget.setAlignment(Qt.AlignCenter)
        self._internal_widget.setMinimumHeight(WIDGET_MIN_HEIGHT)
        self._internal_widget.setWordWrap(True)
        widget = add_unit_label(self.proxy, self._internal_widget,
                                parent=parent)
        widget.setFrameStyle(QFrame.Box | QFrame.Plain)

        action = QAction('Change expression...', widget)
        widget.addAction(action)
        action.triggered.connect(self._change_expression)

        objectName = generateObjectName(self)
        self._style_sheet = (f"QWidget#{objectName}" +
                             " {{ background-color : rgba{}; }}")
        widget.setObjectName(objectName)
        sheet = self._style_sheet.format(ALL_OK_COLOR)
        widget.setStyleSheet(sheet)

        # Add an action for formatting options
        format_action = QAction("Format field..", widget)
        format_action.triggered.connect(self._format_field)
        widget.addAction(format_action)
        self._apply_format(widget)

        return widget

    def clear_widget(self):
        """Clear the internal widget when the device goes offline"""
        self._internal_widget.clear()

    def binding_update(self, proxy):
        self.widget.update_unit_label(proxy)

    def value_update(self, proxy):
        value = get_binding_value(proxy)
        if value is None:
            return

        try:
            disp_value = str(self.function(value))
        except Exception as e:
            disp_value = traceback.format_exception_only(type(e), e)[0]
        self._internal_widget.setText(disp_value)

    def _change_expression(self):
        text, ok = QInputDialog.getText(self.widget, 'Enter Expression',
                                        'f(x) = ', text=self.model.expression)
        if not ok:
            return

        try:
            self.function = eval(f'lambda x: {text}', self.globals_ns)
        except Exception:
            msg = f'Error in expression: {traceback.format_exc()}'
            messagebox.show_warning(msg, title='Error in expression',
                                    parent=self.widget)
            return

        self.model.expression = text
        if get_binding_value(self.proxy) is not None:
            self.value_update(self.proxy)

    # -----------------------------------------------------------------------
    # Formatting methods

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(font_size=dialog.font_size,
                                 font_weight=dialog.font_weight)
            self._apply_format()

    def _apply_format(self, widget=None):
        """The widget is passed as an argument in create_widget as it is not
           yet bound to self.widget then"""
        if widget is None:
            widget = self.widget

        # Apply font formatting
        font = widget.font()
        font.setPointSize(get_font_size_from_dpi(self.model.font_size))
        font.setBold(self.model.font_weight == "bold")
        widget.setFont(font)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import getpass
from functools import partial
from struct import calcsize, pack, unpack

from qtpy.QtCore import QByteArray, QObject, Signal, Slot
from qtpy.QtNetwork import QAbstractSocket, QTcpSocket
from qtpy.QtWidgets import QDialog, QMessageBox, qApp

import karabogui.access as krb_access
from karabo.common.api import (
    KARABO_CONFIG_MANAGER, KARABO_LOGBOOK_MANAGER, KARABO_PROJECT_MANAGER)
from karabo.common.project.api import PROJECT_DB_SCHEMA
from karabo.native import (
    AccessLevel, Hash, Timestamp, decodeBinary, dictToHash, encodeBinary)
from karabogui import background, const
from karabogui.const import REQUEST_REPLY_TIMEOUT
from karabogui.dialogs.reactive_login_dialog import ReactiveLoginDialog
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.logger import get_logger
from karabogui.singletons.api import get_config
from karabogui.util import process_qt_events

MAX_GUI_SERVER_HISTORY = 5
PROJECT_DB_TIMEOUT = 10

logger = get_logger()


class Network(QObject):
    """The `Network` class is the singleton holding the tcp socket for
    the gui server connection.
    """
    signalServerConnectionChanged = Signal(bool)
    signalReceivedData = Signal(object)
    signalNetworkPerformance = Signal(float, bool)

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._tcp_socket = None
        self._data_reader = None
        self._request_queue = []

        self._waiting_messages = {}
        self._show_proc_delay = False

        self.gui_servers = []
        self.hostname = "localhost"
        self.port = "44444"
        self.password = "karabo"
        self.access_level = "observer"

        # Check default settings stored in QSettings!
        self._load_login_settings()

        # Listen for the quit notification
        qApp.aboutToQuit.connect(self.onQuitApplication)

        if get_config()["development"]:
            self.togglePerformanceMonitor()

    def connectToServer(self, parent=None) -> bool:
        """Connection to server via reactive LoginDialog.

        Returns: Boolean to indicate if the login dialog has been accepted
        and that the user has been successfully authenticated.
        """
        dialog = ReactiveLoginDialog(
            access_level=self.access_level,
            hostname=self.hostname,
            port=self.port,
            gui_servers=self.gui_servers,
            parent=parent)

        if dialog.exec() == QDialog.Accepted:
            self.hostname = dialog.hostname
            self.port = dialog.port
            self.gui_servers = dialog.gui_servers
            self.access_level = dialog.access_level
            self.startServerConnection()
            return True

        # Note: The server connection is not changed and thus we must
        # notify the main window to untoggle the button
        self.signalServerConnectionChanged.emit(False)

        return False

    def connectToServerDirectly(self, hostname, port):
        """Connection to server directly via username, host and port

        This function is used for fast startup via the karabo cinema
        """
        self.hostname = hostname
        self.port = port
        dialog = ReactiveLoginDialog(hostname=hostname,
                                     gui_servers=self.gui_servers,
                                     port=port)
        if dialog.exec() != QDialog.Accepted:
            return False
        self.hostname = dialog.hostname
        self.port = dialog.port
        self.gui_servers = dialog.gui_servers
        self.access_level = dialog.access_level
        self.startServerConnection()
        # Allow external runner to see the status of the connection!
        return True

    def disconnectFromServer(self):
        """Disconnect from server"""
        krb_access.reset_login()
        self.signalServerConnectionChanged.emit(False)
        # All panels need to be reset and all projects closed
        process_qt_events(timeout=5000)
        self.endServerConnection()

    def startServerConnection(self):
        """Create the tcp socket and connect to hostname and port"""
        self._tcp_socket = QTcpSocket(self)
        self._tcp_socket.connected.connect(self.onConnected)
        self._tcp_socket.readyRead.connect(self.onReadServerData)
        self._tcp_socket.disconnected.connect(self.onDisconnected)
        self._tcp_socket.error.connect(self.onSocketError)
        self._tcp_socket.connectToHost(self.hostname, self.port)

    def endServerConnection(self):
        """End connection to server and database"""
        self._request_queue = []

        if self._tcp_socket is None:
            return

        self._tcp_socket.disconnectFromHost()
        if (self._tcp_socket.state() == QAbstractSocket.UnconnectedState or
                self._tcp_socket.waitForDisconnected(5000)):
            get_logger().info(
                "Disconnected from the gui server "
                f"<b>{self.hostname}:{self.port}</b>")
            return

        get_logger().error(
            f"Disconnect failed: {self._tcp_socket.errorString()}")

    def togglePerformanceMonitor(self):
        """External method to toggle the performance monitor"""
        self._show_proc_delay = not self._show_proc_delay

    @Slot()
    def onReadServerData(self):
        """Run the network reader generator until it yields"""
        # self._data_reader is a generator object from self._network_generator.
        # Stepping it with next() causes network data to be read.
        next(self._data_reader)

    def _network_generator(self):
        """A generator which continuously reads GUI server messages

        The generator yields when there isn't enough buffered data to be read.
        """
        sizeFormat = 'I'  # unsigned int
        bytesNeededSize = calcsize(sizeFormat)

        while True:
            # Read the size of the Hash
            while self._tcp_socket.bytesAvailable() < bytesNeededSize:
                yield
            rawBytesNeeded = self._tcp_socket.read(bytesNeededSize)
            bytesNeeded = unpack(sizeFormat, rawBytesNeeded)[0]

            # Read the Hash
            while self._tcp_socket.bytesAvailable() < bytesNeeded:
                yield
            dataBytes = self._tcp_socket.read(bytesNeeded)

            # Do something with the Hash at some point in the future.
            self._waiting_messages[id(dataBytes)] = Timestamp().toTimestamp()
            background.executeLater(partial(self.parseInput, dataBytes),
                                    background.Priority.NETWORK)

    def parseInput(self, data):
        """parse the data and emit the signalReceivedData"""
        self._performance_monitor(self._waiting_messages.pop(id(data)))
        self.signalReceivedData.emit(decodeBinary(data))

    @Slot(QAbstractSocket.SocketError)
    def onSocketError(self, socketError):
        print("onSocketError", self._tcp_socket.errorString(), socketError)

        self.disconnectFromServer()

        if socketError == QAbstractSocket.ConnectionRefusedError:
            msg = ('The connection to GUI server <b>{}[:{}]</b> '
                   'failed.').format(self.hostname, self.port)
            reply = QMessageBox.question(
                None, 'Server connection refused', msg,
                QMessageBox.Retry | QMessageBox.Cancel, QMessageBox.Retry)

            if reply == QMessageBox.Cancel:
                return
        elif socketError == QAbstractSocket.RemoteHostClosedError:
            msg = ('The remote host <b>{}[:{}]</b> closed the '
                   'connection').format(self.hostname, self.port)
            reply = QMessageBox.question(
                None, 'Connection closed', msg,
                QMessageBox.Retry | QMessageBox.Cancel, QMessageBox.Retry)

            if reply == QMessageBox.Cancel:
                return
        elif socketError == QAbstractSocket.HostNotFoundError:
            msg = 'The host address <b>{}</b> was not found.'.format(
                self.hostname)
            reply = QMessageBox.question(
                None, 'Host address error', msg,
                QMessageBox.Retry | QMessageBox.Cancel, QMessageBox.Retry)

            if reply == QMessageBox.Cancel:
                return
        elif socketError == QAbstractSocket.NetworkError:
            msg = ('An error occurred with the network (e.g., <br>the network '
                   'cable was accidentally plugged out).')
            reply = QMessageBox.question(
                None, 'Network error', msg,
                QMessageBox.Retry | QMessageBox.Cancel, QMessageBox.Retry)

            if reply == QMessageBox.Cancel:
                return
        elif socketError == QAbstractSocket.SocketAccessError:
            msg = ('The socket operation failed because the application '
                   'lacked the required privileges.')
            reply = QMessageBox.question(
                None, 'Network error', msg,
                QMessageBox.Retry | QMessageBox.Cancel, QMessageBox.Retry)

            if reply == QMessageBox.Cancel:
                return
        else:
            msg = ('An unknown socket operation error occured. '
                   'Type: {} - Description: {}').format(
                socketError, self._tcp_socket.errorString())
            reply = QMessageBox.question(
                None, 'Network error', msg,
                QMessageBox.Retry | QMessageBox.Cancel, QMessageBox.Retry)
            if reply == QMessageBox.Cancel:
                return

        self.connectToServer()

    def onServerConnection(self, connect, parent=None):
        """Connect or disconnect depending on the input parameter

        This function is called from the main window directly!

        :param connect: Either True or False
        :type connect: bool
        """
        if connect:
            self.connectToServer(parent)
        else:
            self.disconnectFromServer()

    @Slot()
    def onQuitApplication(self):
        """This slot is triggered from the MainWindow

        The user wants to quit the application so the network connection
        needs to be closed and the manager needs to be informed to close
        the database connection as well.
        """
        self.endServerConnection()

    @Slot()
    def onConnected(self):
        def _least_recently_used(item, sequence, maxsize):
            """Apply of the Least Recently Used (LRU) algorithm on sequence

            The LRU algorithm evicts from sequence the least recently used
            element and item is moved to the sequence front.
            As the result, sequence always contains its most popular element
            """
            if item in sequence:
                sequence.remove(item)
            sequence = [item] + sequence
            return sequence if len(sequence) <= maxsize else sequence[:maxsize]

        # cache the server address
        server = f'{self.hostname}:{self.port}'

        self.gui_servers = _least_recently_used(server, self.gui_servers,
                                                int(MAX_GUI_SERVER_HISTORY))

        # Save to singleton!
        get_config()['access_level'] = self.access_level
        get_config()['gui_servers'] = self.gui_servers
        self._data_reader = self._network_generator()

        # Update MainWindow toolbar
        self.signalServerConnectionChanged.emit(True)

    @Slot()
    def onDisconnected(self):
        """The tcp socket was disconnected"""

    # ---------------------------------------------------------------------
    # Protocol methods

    def onKillDevice(self, device_id):
        logger.info(f"Sending request to shutdown device <b>{device_id}</b>")
        h = Hash("type", "killDevice")
        h["deviceId"] = device_id
        self._write_hash(h)

    def onKillServer(self, server_id):
        logger.info(f"Sending request to shutdown server <b>{server_id}</b>")
        h = Hash("type", "killServer")
        h["serverId"] = server_id
        self._write_hash(h)

    def onGetDeviceConfiguration(self, device_id):
        h = Hash("type", "getDeviceConfiguration")
        h["deviceId"] = device_id
        self._write_hash(h)

    def onReconfigure(self, device_id, configuration):
        """Set values in a device
        """
        props = ", ".join(configuration.keys())
        logger.info(f"Request to reconfigure the properties <b>{props}</b> "
                    f" of device <b>{device_id}</b>")
        h = Hash("type", "reconfigure")
        h["deviceId"] = device_id
        h["configuration"] = configuration
        h["reply"] = True
        h["timeout"] = REQUEST_REPLY_TIMEOUT
        self._write_hash(h)

    def onInitDevice(self, server_id, class_id, device_id, config,
                     attrUpdates=None):
        logger.info(f"Requesting to instantiate device <b>{device_id}</b> "
                    f"with <b>{class_id}</b> on server <b>{server_id}</b>")
        h = Hash("type", "initDevice")
        h["serverId"] = server_id
        h["classId"] = class_id
        h["deviceId"] = device_id
        h["configuration"] = config
        if attrUpdates is not None:
            h["schemaUpdates"] = attrUpdates
        self._write_hash(h)

    def onExecute(self, device_id, slot_name, ignore_timeouts):
        logger.info(f"Executing slot <b>{slot_name}</b> "
                    f"of device <b>{device_id}</b>")
        h = Hash("type", "execute")
        h["deviceId"] = device_id
        h["command"] = slot_name
        h["reply"] = True
        if not ignore_timeouts:
            h["timeout"] = REQUEST_REPLY_TIMEOUT
        self._write_hash(h)

    def onExecuteGeneric(self, instanceId, slot_name, params, token=None):
        logger.info(f"Executing slot <b>{slot_name}</b> "
                    f"of instance <b>{instanceId}</b>")
        h = Hash("type", "requestGeneric")
        if token is not None:
            h["token"] = token
        h["instanceId"] = instanceId
        h["slot"] = slot_name
        h["args"] = params
        h["timeout"] = REQUEST_REPLY_TIMEOUT
        h["replyType"] = "requestGeneric"
        self._write_hash(h)

    def onStartMonitoringDevice(self, device_id):
        h = Hash("type", "startMonitoringDevice")
        h["deviceId"] = device_id
        self._write_hash(h)

    def onStopMonitoringDevice(self, device_id):
        h = Hash("type", "stopMonitoringDevice")
        h["deviceId"] = device_id
        self._write_hash(h)

    def onGetClassSchema(self, server_id, class_id):
        h = Hash("type", "getClassSchema")
        h["serverId"] = server_id
        h["classId"] = class_id
        self._write_hash(h)

    def onGetDeviceSchema(self, device_id):
        h = Hash("type", "getDeviceSchema")
        h["deviceId"] = device_id
        self._write_hash(h)

    def onGetPropertyHistory(self, device_id, path, t0, t1, maxNumData):
        h = Hash("type", "getPropertyHistory")
        h["deviceId"] = device_id
        h["property"] = path
        h["t0"] = t0
        h["t1"] = t1
        h["maxNumData"] = maxNumData
        self._write_hash(h)

    # ---------------------------------------------------------------------
    # Current Configuration Interface

    def onGetConfigurationFromPast(self, device_id, time, preview):
        logger.info(f"Requesting configuration for device "
                    f"<b>{device_id}</b> at time point {time}")
        h = Hash("type", "getConfigurationFromPast")
        h["deviceId"] = device_id
        h["time"] = time
        h["preview"] = preview
        self._write_hash(h)

    def onGetInitConfiguration(self, device_id, name, preview):
        logger.info(f"Requesting named configuration <b>{name}</b> "
                    f"for device <b>{device_id}</b>")
        h = Hash("type", "requestGeneric")
        args = Hash(
            "deviceId", device_id,
            "name", name)
        h["args"] = args
        h["timeout"] = REQUEST_REPLY_TIMEOUT
        h["instanceId"] = KARABO_CONFIG_MANAGER
        h["preview"] = preview
        h["slot"] = "slotGetInitConfiguration"
        h["replyType"] = "getInitConfiguration"
        self._write_hash(h)

    def onListInitConfigurations(self, device_id, name=""):
        logger.info("Requesting list of named configurations for "
                    f"device <b>{device_id}</b>")
        h = Hash("type", "requestGeneric")
        args = Hash(
            "name", name,
            "deviceId", device_id)
        h["args"] = args
        h["timeout"] = REQUEST_REPLY_TIMEOUT
        h["instanceId"] = KARABO_CONFIG_MANAGER
        h["slot"] = "slotListInitConfigurations"
        h["replyType"] = "listInitConfigurations"
        self._write_hash(h)

    def onDeleteInitConfiguration(self, device_id, name):
        logger.info("Requesting delete of named configurations for "
                    f"device <b>{device_id}:{name}</b>")
        h = Hash("type", "requestGeneric")
        args = Hash(
            "name", name,
            "deviceId", device_id)
        h["args"] = args
        h["timeout"] = REQUEST_REPLY_TIMEOUT
        h["instanceId"] = KARABO_CONFIG_MANAGER
        h["slot"] = "slotDeleteInitConfiguration"
        h["replyType"] = "deleteInitConfiguration"
        self._write_hash(h)

    def onSaveInitConfiguration(self, name, deviceIds, update=False):
        logger.info(f"Saving configuration by name {name} for devices "
                    f"<b>{deviceIds}</b>")
        h = Hash("type", "requestGeneric")
        args = Hash(
            "name", name,
            "deviceIds", deviceIds,
            "client", const.KARABO_CLIENT_ID)
        h["args"] = args
        h["update"] = update
        h["timeout"] = REQUEST_REPLY_TIMEOUT
        h["instanceId"] = KARABO_CONFIG_MANAGER
        h["slot"] = "slotSaveInitConfiguration"
        h["replyType"] = "saveInitConfiguration"
        self._write_hash(h)

    # ---------------------------------------------------------------------
    # Current Project Interface

    def _request_project_message(self, action_type, reply_type, timeout=None,
                                 **kwargs):
        h = Hash("type", "requestGeneric")
        args = Hash("type", action_type)
        for key, value in kwargs.items():
            args[key] = value
        h["args"] = args
        h["empty"] = True
        h["timeout"] = timeout or PROJECT_DB_TIMEOUT
        h["instanceId"] = KARABO_PROJECT_MANAGER
        h["slot"] = "slotGenericRequest"
        h["replyType"] = reply_type
        self._write_hash(h)

    def onListProjectDomains(self, project_manager):
        self._request_project_message(
            "listDomains", "projectListDomains")

    def onProjectListItems(self, project_manager, domain, item_type):
        self._request_project_message("listItems", "projectListItems",
                                      domain=domain, item_types=[item_type])

    def onProjectListProjectsWithDevice(self, project_manager,
                                        domain, device_id):
        self._request_project_message(
            "listProjectsWithDevice", "projectListProjectsWithItem",
            timeout=120, name=device_id, domain=domain)

    def onProjectListProjectsWithMacro(self, project_manager,
                                       domain, macro_id):
        self._request_project_message(
            "listProjectsWithMacro", "projectListProjectsWithItem",
            timeout=120, name=macro_id, domain=domain)

    def onProjectListProjectsWithServer(self, project_manager,
                                        domain, server_id):
        self._request_project_message(
            "listProjectsWithServer", "projectListProjectsWithItem",
            timeout=120, name=server_id, domain=domain)

    def onProjectLoadItems(self, project_manager, items):
        self._request_project_message(
            "loadItems", "projectLoadItems", timeout=30, items=items)

    def onProjectSaveItems(self, project_manager, items):
        self._request_project_message(
            "saveItems", "projectSaveItems", timeout=30,
            client=const.KARABO_CLIENT_ID, items=items,
            schema_version=PROJECT_DB_SCHEMA)

    def onProjectUpdateTrashed(self, project_manager, items):
        self._request_project_message("updateTrashed",
                                      "projectUpdateTrashed", items=items)

    # ---------------------------------------------------------------------

    def onAcknowledgeAlarm(self, instanceId, rowId):
        logger.info(f"Acknowledging alarm with ID <b>{rowId}</b> "
                    f"of alarm service <b>{instanceId}</b>")
        h = Hash("type", "acknowledgeAlarm")
        h["alarmInstanceId"] = instanceId
        h["acknowledgedRows"] = Hash(rowId, True)
        self._write_hash(h)

    def onRequestAlarms(self, instanceId):
        h = Hash("type", "requestAlarms")
        h["alarmInstanceId"] = instanceId
        self._write_hash(h)

    def onSubscribeToOutput(self, device_id, path, subscribe):
        h = Hash("type", "subscribeNetwork")
        h["channelName"] = device_id + ":" + path
        h["subscribe"] = subscribe
        self._write_hash(h)

    def onRequestNetwork(self, name):
        h = Hash("type", "requestNetwork", "channelName", name)
        self._write_hash(h)

    def onInfo(self, info: Hash):
        # XXX: Change `type` to `info` in Karabo 3
        assert isinstance(info, Hash)
        h = Hash("type", "error", "info", info,
                 "clientId", const.KARABO_CLIENT_ID)
        self._write_hash(h)

    def onGetGuiSessionInfo(self):
        h = Hash("type", "getGuiSessionInfo",
                 "clientId", const.KARABO_CLIENT_ID)
        self._write_hash(h)

    def onSetLogLevel(self, instanceId, level):
        h = Hash("type", "setLogLevel", "instanceId", instanceId,
                 "level", level)
        self._write_hash(h)

    # ------------------------------------------------------------------------

    def set_server_information(self, read_only=False, **kwargs):
        """We get the reply from the GUI Server and set the information"""
        if read_only:
            default = AccessLevel.OBSERVER
        else:
            default = AccessLevel(
                krb_access.ACCESS_LEVEL_MAP.get(self.access_level,
                                                AccessLevel.EXPERT))

        krb_access.GLOBAL_ACCESS_LEVEL = default
        # Inform the GUI to change correspondingly the allowed
        # level-downgrade
        broadcast_event(KaraboEvent.LoginUserChanged, {})
        self.onLogin()
        self._empty_request_queue()

    # ------------------------------------------------------------------------
    # private functions

    def _load_login_settings(self):
        """Load the login settings from the configuration singleton

        This method sets the default `host`, `port` and `gui_servers`
        """
        config = get_config()
        self.access_level = config['access_level']
        self.gui_servers = config['gui_servers']

        if self.gui_servers:
            self.hostname, self.port = self.gui_servers[0].split(':')

        self.port = int(self.port)

    def _write_hash(self, h):
        # There might be a connect to server in progress, but without success
        if (self._tcp_socket is None or
                self._tcp_socket.state() == QAbstractSocket.HostLookupState or
                self._tcp_socket.state() == QAbstractSocket.ConnectingState):
            # Save request for connection established
            self._request_queue.append(h)
            return

        stream = QByteArray()
        dataBytes = encodeBinary(h)
        stream.push_back(QByteArray(pack('I', len(dataBytes))))
        stream.push_back(dataBytes)
        self._tcp_socket.write(stream)

    def _performance_monitor(self, received_timestamp):
        diff = Timestamp().toTimestamp() - received_timestamp
        self.signalNetworkPerformance.emit(diff, self._show_proc_delay)

    def onLogin(self):
        login_info = Hash("type", "login")
        # username to transport ClientID is deprecated - remove it after a
        # couple of new versions of the GUI Server.
        login_info["username"] = const.KARABO_CLIENT_ID
        # Redundantly send, during the above key deprecation period, the
        # the ClientID with its new key.
        login_info["clientId"] = const.KARABO_CLIENT_ID
        login_info["version"] = const.GUI_VERSION
        login_info["applicationMode"] = const.APPLICATION_MODE
        if krb_access.is_authenticated():
            login_info["oneTimeToken"] = krb_access.ONE_TIME_TOKEN
            # For authenticated logins, don't set the "clientUserId" - the
            # authenticated userId and the authorized AccessLevel will
            # be returned later as a result of the token validation.
        else:
            # Non authenticated logins:
            # Send the OS id of the user running the process to the server for
            # logging/auditing purposes.
            login_info["clientUserId"] = getpass.getuser()

        login_info["info"] = dictToHash(get_config().info())
        self._write_hash(login_info)

    def beginTemporarySession(self, **info):
        h = Hash("type", "beginTemporarySession")
        h["clientId"] = const.KARABO_CLIENT_ID
        h["version"] = const.GUI_VERSION
        h["temporarySessionToken"] = info["temporarySessionToken"]
        h["levelBeforeTemporarySession"] = info["levelBeforeTemporarySession"]
        self._write_hash(h)

    def endTemporarySession(self, **info):
        h = Hash("type", "endTemporarySession")
        h["temporarySessionToken"] = krb_access.ONE_TIME_TOKEN
        self._write_hash(h)

    def _empty_request_queue(self):
        """If some requests got piled up, because of no server connection,
        now these get handled after the login"""
        for r in self._request_queue:
            self._write_hash(r)
        self._request_queue = []

    def listDestinations(self):
        logger.info("Fetching available active proposals in the Topic.")
        h = Hash("type", "requestGeneric")
        h["instanceId"] = KARABO_LOGBOOK_MANAGER
        h["slot"] = "slotListDestinations"
        h["replyType"] = "listDestinations"
        h["args"] = Hash()
        self._write_hash(h)

    def onSaveLogBook(self, **info):
        h = Hash()
        h["type"] = "requestGeneric"
        h["args"] = Hash(info)
        h["instanceId"] = KARABO_LOGBOOK_MANAGER
        h["slot"] = "slotSaveLogBook"
        h["replyType"] = "saveLogBook"
        self._write_hash(h)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QPoint, QRectF
from qtpy.QtWidgets import QBoxLayout

from karabo.common.scenemodel.api import (
    BoxLayoutModel, DisplayLabelModel, LabelModel, SceneModel)
from karabo.common.scenemodel.tests.utils import single_model_round_trip
from karabogui.sceneview.api import SceneView
from karabogui.sceneview.tools.clipboard import (
    SceneAlignAction, SceneMoveAction, _add_models_to_clipboard,
    _read_models_from_clipboard)
from karabogui.testing import lazy_scene_reading


def test_move_action(gui_app):
    label = LabelModel(x=10, y=10, text="bar")
    labels = [label]
    model = SceneModel(children=labels)
    scene_view = SceneView(model=model)
    selection_model = scene_view.selection_model
    widget = scene_view.widget_at_position(QPoint(10, 10))
    selection_model.select_object(widget)

    # Left
    action = SceneMoveAction(text="Left")
    action.perform(scene_view)
    assert label.x == 0
    assert label.y == 10

    # Right
    action = SceneMoveAction(text="Right")
    action.perform(scene_view)
    assert label.x == 10
    assert label.y == 10

    # Up
    action = SceneMoveAction(text="Up")
    action.perform(scene_view)
    assert label.x == 10
    assert label.y == 0

    # Down
    action = SceneMoveAction(text="Down")
    action.perform(scene_view)
    assert label.x == 10
    assert label.y == 10

    # Without snap grid, simple pixel movement
    scene_view.snap_to_grid = False

    # Left
    action = SceneMoveAction(text="Left")
    action.perform(scene_view)
    assert label.x == 9
    assert label.y == 10

    # Right
    action = SceneMoveAction(text="Right")
    action.perform(scene_view)
    assert label.x == 10
    assert label.y == 10

    # Up
    action = SceneMoveAction(text="Up")
    action.perform(scene_view)
    assert label.x == 10
    assert label.y == 9

    # Down
    action = SceneMoveAction(text="Down")
    action.perform(scene_view)
    assert label.x == 10
    assert label.y == 10


def test_align_action(gui_app):
    for action in ("Left", "Right", "Top", "Bottom"):
        labels = [
            LabelModel(x=i * 10, y=i * 10, text="bar") for i in range(4)
        ]
        model = SceneModel(children=labels)
        scene_view = SceneView(model=model)
        scene_view.select_all()

        max_right = max(label.x + label.width for label in labels)
        max_down = max(label.x + label.width for label in labels)

        action = SceneAlignAction(text=action)
        action.perform(scene_view)
        for label in labels:
            if action == "Left":
                assert label.x == 0
            elif action == "Right":
                assert label.x + label.width == max_right
            if action == "Top":
                assert label.y == 0
            elif action == "Bottom":
                assert label.y + label.height == max_down


def test_copy_paste(gui_app):
    """These are standalone widgets that are not bound to any devices."""
    # Test single model selection
    with lazy_scene_reading():
        label_model = _get_label_model("foo")
        _assert_copy_paste(label_model)

        # Test single layout selection
        layout_model = _get_layout_model(label_model)
        _assert_copy_paste(layout_model)

        # Test multiple models selection
        another_label = _get_label_model("bar", x=10, y=10)
        _assert_copy_paste(label_model, another_label)

        # Test multiple layouts selection
        another_layout = _get_layout_model(another_label, x=10, y=10)
        _assert_copy_paste(layout_model, another_layout)


def _assert_copy_paste(*models):
    # Do a single round trip for each models to verify their integrity
    read_models = [single_model_round_trip(model) for model in models]
    for read_model, model in zip(read_models, models):
        _assert_model(read_model, model)

    # Mock copy to clipboard
    _add_models_to_clipboard(read_models, rect=QRectF(0, 0, 0, 0))
    copied_models = _read_models_from_clipboard()
    assert len(copied_models) == len(models)

    # Checked copied model has same traits as read model
    for copied_model, read_model in zip(copied_models, read_models):
        _assert_model(copied_model, read_model)


def _assert_model(actual, expected):
    trait_names = expected.copyable_trait_names()
    if "children" in trait_names:
        trait_names.remove("children")
        iter_children = zip(actual.children, expected.children)
        for actual_child, expected_child in iter_children:
            _assert_model(actual_child, expected_child)

    assert type(actual) is type(expected)
    assert actual.trait_get(trait_names) == expected.trait_get(trait_names)


def _get_label_model(text, x=0, y=0):
    UBUNTU_FONT_SPEC = 'Ubuntu,48,-1,5,63,0,0,0,0,0'
    traits = {'x': x, 'y': y, 'height': 100, 'width': 100,
              'text': text, 'font': UBUNTU_FONT_SPEC,
              'foreground': '#000000', 'background': '#ffffff',
              'frame_width': 0}
    return LabelModel(**traits)


def _get_layout_model(model, x=0, y=0):
    # Initialize children models
    height = 30
    label_model = LabelModel(width=100, height=height, text="bar")
    display_model = DisplayLabelModel(width=200, height=height)
    model.trait_set(width=300, height=height)
    children_model = [label_model, display_model, model]

    # Initialize layout model
    hbox_model = BoxLayoutModel(x=x, y=y, children=children_model,
                                direction=QBoxLayout.LeftToRight)

    # Correct the position from the initial position and width
    x0, y0 = (x, y)
    for child in children_model:
        child.trait_set(x=x0, y=y0)
        # Add child width for the next iteration
        x0 += child.width
    hbox_model.height = height
    hbox_model.width = sum([child.width for child in children_model])
    return hbox_model



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy
from pytest import raises as assert_raises

from karabo.native import AccessLevel, AccessMode, Assignment, Hash
from karabogui.binding.api import (
    BindingNamespace, BoolBinding, ByteArrayBinding, CharBinding, FloatBinding,
    HashBinding, Int8Binding, Int16Binding, Int32Binding, Int64Binding,
    NDArrayBinding, NodeBinding, NoneBinding, SchemaBinding, SlotBinding,
    StringBinding, Uint8Binding, Uint16Binding, Uint32Binding, Uint64Binding,
    VectorBoolBinding, VectorCharBinding, VectorDoubleBinding,
    VectorFloatBinding, VectorHashBinding, VectorInt8Binding,
    VectorInt16Binding, VectorInt32Binding, VectorInt64Binding,
    VectorNoneBinding, VectorStringBinding, VectorUint8Binding,
    VectorUint16Binding, VectorUint32Binding, VectorUint64Binding)

ALL_BINDINGS = (
    BoolBinding, ByteArrayBinding, CharBinding,
    FloatBinding, HashBinding, Int8Binding, Int16Binding,
    Int32Binding, Int64Binding, NDArrayBinding, NodeBinding, NoneBinding,
    SchemaBinding, SlotBinding, StringBinding, Uint8Binding, Uint16Binding,
    Uint32Binding, Uint64Binding, VectorBoolBinding, VectorCharBinding,
    VectorDoubleBinding, VectorFloatBinding, VectorHashBinding,
    VectorInt8Binding, VectorInt16Binding, VectorInt32Binding,
    VectorInt64Binding, VectorNoneBinding, VectorStringBinding,
    VectorUint8Binding, VectorUint16Binding, VectorUint32Binding,
    VectorUint64Binding
)


def test_binding_traits():
    # Defined trait names in Binding models,
    # children_names should not be used for binding types whose value are of
    # type BindingNamespace
    expected_traits = ('_attributes', 'accessMode', 'assignment',
                       'attributes', 'children_names', 'options',
                       'requiredAccessLevel', 'timestamp', 'unit_label',
                       'value', 'displayedName', 'displayType', 'row_schema')
    for binding_factory in ALL_BINDINGS:
        obj = binding_factory()
        for trait in obj.copyable_trait_names():
            assert trait in expected_traits

        assert obj.requiredAccessLevel is AccessLevel.OBSERVER
        assert obj.displayedName == ""
        assert obj.displayType == ""
        assert obj.accessMode is AccessMode.UNDEFINED
        assert obj.assignment is Assignment.OPTIONAL
        if isinstance(obj.options, numpy.ndarray):
            # truth value of array ambiguous, must check size explicitly
            assert obj.options.size == 0
        else:
            assert not obj.options
        assert obj.unit_label == ""


def test_binding_namespace():
    ns = BindingNamespace()

    assert list(ns) == []

    ns.foo = 0
    ns.bar = 'baz'
    assert list(ns) == ['foo', 'bar']

    del ns.bar
    assert list(ns) == ['foo']
    assert repr(ns) == "{\n\tfoo: 0\n}"

    ns.clear_namespace()
    assert list(ns) == []

    with assert_raises(AttributeError):
        del ns.foo

    ints_only = BindingNamespace(item_type=int)
    with assert_raises(ValueError):
        ints_only.not_an_int = 3.14


def test_table_binding():
    vector_hash = VectorHashBinding()
    assert vector_hash.row_schema == Hash()
    vector_hash.update_attributes({})
    assert vector_hash.row_schema == Hash()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import contextlib

from qtpy.QtCore import QAbstractItemModel, QModelIndex
from traits.api import HasStrictTraits, WeakRef


class _UpdateContext(HasStrictTraits):
    """A context manager that can be handed off to code which doesn't need to
    know that it's dealing with a Qt QAbstractItemModel.
    """
    item_model = WeakRef(QAbstractItemModel)

    @contextlib.contextmanager
    def reset_context(self):
        """Provide a context whenever the system tree was cleared then the
        Qt model needs to be reseted.

        NOTE: This method is a context manager wraps the insertion with calls
        to ``QAbstractItemModel.beginResetModel`` and
        ``QAbstractItemModel.endResetModel`` (See Qt documentation)
        """
        try:
            self.item_model.beginResetModel()
            yield
        finally:
            self.item_model.endResetModel()

    @contextlib.contextmanager
    def insertion_context(self, parent_node, first, last):
        """Provide a context for the addition of multiple children under a
        single parent item.

        NOTE: This method is a context manager wraps the insertion with calls
        to ``QAbstractItemModel.beginInsertRows`` and
        ``QAbstractItemModel.endInsertRows`` (See Qt documentation)
        """
        parent_index = self.item_model.createIndex(parent_node.row(), 0,
                                                   parent_node)

        def gen():
            try:
                self.item_model.beginInsertRows(parent_index, first, last)
                yield
            finally:
                self.item_model.endInsertRows()

        if parent_index.isValid():
            yield from gen()
        else:
            yield

    @contextlib.contextmanager
    def removal_context(self, tree_node):
        """Provide a context for the removal of a single item from the model.

        NOTE: This method is a context manager which wraps the removal of an
        item with ``QAbstractItemModel.beginRemoveRows`` and
        ``QAbstractItemModel.endRemoveRows`` (See Qt documentation)
        """
        node_row = tree_node.row()
        index = self.item_model.createIndex(node_row, 0, tree_node)
        if index.isValid():
            parent_index = index.parent()
        else:
            parent_index = QModelIndex()

        def gen():
            try:
                self.item_model.beginRemoveRows(parent_index, node_row,
                                                node_row)
                yield
            finally:
                self.item_model.endRemoveRows()

        if parent_index.isValid():
            yield from gen()
        else:
            yield

    @contextlib.contextmanager
    def insert_root_context(self, first, last):
        """Provide a context for the addition of a children under the root"""
        try:
            self.item_model.beginInsertRows(QModelIndex(), first, last)
            yield
        finally:
            self.item_model.endInsertRows()

    @contextlib.contextmanager
    def remove_root_context(self, first, last):
        """Provide a context for the removal of a children under the root"""
        try:
            self.item_model.beginRemoveRows(QModelIndex(), first, last)
            yield
        finally:
            self.item_model.endRemoveRows()

    @contextlib.contextmanager
    def removal_children_context(self, tree_node):
        """Provide a context for the removal of a single item from the model.
        """
        node_row = tree_node.row()
        index = self.item_model.createIndex(node_row, 0, tree_node)

        def gen():
            try:
                self.item_model.beginRemoveRows(index, 0,
                                                len(tree_node.children) - 1)
                yield
            finally:
                self.item_model.endRemoveRows()

        if index.isValid() and tree_node.children:
            yield from gen()
        else:
            yield

    @contextlib.contextmanager
    def layout_context(self):
        """Provide a context for layout change announcements
        """
        try:
            self.item_model.layoutAboutToBeChanged.emit(
                [], QAbstractItemModel.VerticalSortHint)
            yield
        finally:
            self.item_model.layoutChanged.emit()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import operator
import os
import subprocess
import sys
from platform import system
from urllib.parse import parse_qs, urlparse

GUI = "gui"
CINEMA = "cinema"
THEATRE = "theatre"

KARABO_GUI = "karabogui.programs.gui_runner"
KARABO_CINEMA = "karabogui.programs.cinema"
KARABO_THEATRE = "karabogui.programs.theatre"

KEEP_BLANK_VALUES = True

PYTHON = sys.executable


def parse_url(url):
    parsed = urlparse(url)
    return parsed.hostname, parse_qs(parsed.query,
                                     keep_blank_values=KEEP_BLANK_VALUES)


def get_cinema_args(queries):
    # Check required arguments:
    required = {"domain": lambda x: operator.eq(x, 1),
                "scene_uuid": lambda x: operator.ge(x, 1)}
    _check_args(queries, required)

    domain = queries["domain"][-1].upper()
    # NOTE: `scene_uuid` maps to the `scene_uuids` option in the
    # karabo-cinema executable. This is because the standard way of passing
    # a list of the same option in a url is `something?id=1&id=2`
    # to pass 1 and 2 to the id in the same query.
    scene_uuid = [uuid.lower() for uuid in queries["scene_uuid"]]
    args = [domain, *scene_uuid]

    # Check optional arguments:
    if queries.get("nosplash", False):
        args.append("--nosplash")
    # Check if host and port is both supplied
    if "host" in queries and "port" in queries:
        args.extend(["--host", queries["host"][-1],
                     "--port", queries["port"][-1]])
    return args


def get_theatre_args(queries):
    # Check required arguments:
    required = {"scene_id": lambda x: operator.ge(x, 1)}
    assert _check_args(queries, required)

    # NOTE: `scene_id` maps to the `scene_ids` option in the
    # karabo-theatre executable. This is because the standard way of passing
    # a list of the same option in a url is `something?id=1&id=2`
    # to pass 1 and 2 to the id in the same query.
    args = queries["scene_id"]

    # Check optional arguments:
    if queries.get("nosplash", False):
        args.append("--nosplash")
    # Check if host and port is both supplied
    if "host" in queries and "port" in queries:
        args.extend(["--host", queries["host"][-1],
                     "--port", queries["port"][-1]])
    return args


def _check_args(queries, required=None):
    if required is None:
        # Nothing to do here!
        return True

    invalid = []
    for arg, requirements in required.items():
        default = [''] if KEEP_BLANK_VALUES else None
        value = queries.get(arg, default)
        if value == default or not requirements(len(value)):
            invalid.append(arg)

    if invalid:
        raise ValueError(f"Invalid arguments found: {', '.join(invalid)}")

    return True


def main(url):
    # This is run from the application shortcut
    if not len(url):
        host, queries = None, ()
    else:
        # Assume that there's only one argument from the handler,
        # which is the URL
        assert len(url) == 1, "Invalid URL is supplied."
        url, = url
        host, queries = parse_url(url)

    if host is None or host == GUI:
        # Run GUI
        cmd = [PYTHON, "-m", KARABO_GUI]
    elif host == CINEMA:
        # Run Cinema
        cmd = [PYTHON, "-m", KARABO_CINEMA, *get_cinema_args(queries)]
    elif host == THEATRE:
        # Run Theatre
        cmd = [PYTHON, "-m", KARABO_THEATRE, *get_theatre_args(queries)]
    else:
        # Nothing to do here!
        return

    # Run command!
    if system() == "Windows":
        _run_detached_command_windows(cmd)
    else:
        subprocess.Popen(cmd, preexec_fn=os.setsid)


def _run_detached_command_windows(cmd):
    flags = 0
    flags |= 0x00000008  # DETACHED_PROCESS
    flags |= 0x00000200  # CREATE_NEW_PROCESS_GROUP
    flags |= 0x08000000  # CREATE_NO_WINDOW
    subprocess.Popen(cmd, creationflags=flags, close_fds=True)


if __name__ == "__main__":
    main(sys.argv[1:])



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QWidget
from traits.api import (
    Bool, HasStrictTraits, Instance, List, Property, on_trait_change)

from karabo.common.scenemodel.api import BaseWidgetObjectData
from karabogui import background
from karabogui.binding.api import PropertyProxy, get_binding_value

from .util import get_class_const_trait


class BaseBindingController(HasStrictTraits):
    """The base class of controllers which mediate the connection between a
    data binding model (and a scene model) and a scene widget view.
    """
    # Scene data model describing the widget
    model = Instance(BaseWidgetObjectData, args=())
    # The proxy of the main property being visualized
    proxy = Instance(PropertyProxy)
    # Convenience property for getting all proxies for a widget
    proxies = Property(List, depends_on=['proxy', '_additional_proxies'])
    # The QWidget containing the binding visualization
    widget = Instance(QWidget)

    # Additional proxies added to the widget
    _additional_proxies = List(Instance(PropertyProxy))
    _showing = Bool(False)
    _deferred = Bool(False)

    # -------------------------------------------------------------------------
    # Subclass interface

    def add_proxy(self, proxy):
        """Implemented by subclasses to catch new `PropertyProxy` instances as
        they are added to the controller. Returns `True` if the controller
        can make use of the added proxy instance.

        OPTIONAL: Not all widgets are capable of visualizing mutiple properties
        """
        raise NotImplementedError

    def remove_proxy(self, proxy):
        """Implemented by subclasses to remove `PropertyProxy` instances.
        Returns `True` if the controller can make use of the removed proxy
        instance.

        OPTIONAL: Not all widgets allow removal of properties
        """
        raise NotImplementedError

    def binding_update(self, proxy):
        """Implemented by subclasses to receive notifications that the
        `binding` trait of a proxy attached to the controller has been updated.

        OPTIONAL: Not all widgets care when their bindings change.
        """

    def create_widget(self, parent):
        """Implemented by subclasses to create the widget needed to visualize
        a binding.
        """
        raise NotImplementedError

    @classmethod
    def initialize_model(cls, proxy, model):
        """Implemented by subclasses to initialize a model from a proxy.

        This method is only invoked on controller creation
        """

    def destroy_widget(self):
        """Implemented by subclasses to clean up anything in the widget which
        is otherwise not automatically cleaned up.

        OPTIONAL: Not all widgets will need additional cleanup.
        """

    def deferred_update(self):
        """Implemented by subclasses to update controller as requested by
        `update_later`.

        OPTIONAL: Override this method to run long-running code, and call
        `update_later` instead.
        """

    def set_read_only(self, readonly):
        """Implemented by subclasses to notify a widget of its read-only status

        OPTIONAL: (if can_edit=True passed to register_binding_controller)
        """

    def state_update(self, proxy):
        """Implemented by subclasses to receive notifications that the device
        parent of `proxy` has updated its state.

        OPTIONAL: Not all widgets care about device state
        """

    def value_update(self, proxy):
        """Implemented by subclasses to receive notifications that the
        data value in a proxy attached to the controller has been updated.

        "OPTIONAL": Not all widgets care when their values change, but most do.
        """

    # -------------------------------------------------------------------------
    # Public interface

    def create(self, parent):
        """Create the widget for this controller
        """
        self.widget = self.create_widget(parent)

    def destroy(self):
        """Destroys the widget, giving subclasses a chance to do some cleanup
        """
        self.destroy_widget()

        if self.widget:
            self.widget.setParent(None)
            self.widget = None

        self.hide()
        self.trait_setq(proxy=None, _additional_proxies=[])

    def finish_initialization(self):
        """Force the controller to update itself in the absence of a device
        update on its main proxy.
        """
        if self.widget is None or self.proxy.binding is None:
            return

        # controllers which specify themselves `_can_show_nothing` should
        # be able to deal with Undefined, i.e. their value_update() and
        # state_update() function should use get_binding_value() and
        # specify a proper default value if the proxy.value is Undefined
        self._proxy_update(self.proxy)

    def hide(self):
        """Hide the proxies. Stops monitoring the parent device of each proxy
        being visualized.
        """
        if not self._showing:
            return

        for proxy in self.proxies:
            proxy.stop_monitoring()
        self._showing = False

    def show(self):
        """Show the proxies. Starts monitoring the parent device of each proxy
        being visualized.
        """
        if self._showing:
            return

        for proxy in self.proxies:
            proxy.start_monitoring()
        self._showing = True

    def update_later(self):
        """Call longer-running code at a later time

        If the controller has code that takes a while to run, it should be
        moved to the `deferred_update` method and the widget should call
        `update_later` to schedule this update.

        This keeps the GUI responsive.

        This is especially important for properties which change often, as
        `deferred_update` will only be called after many updates have finished.
        """

        def updater():
            self.deferred_update()
            self._deferred = False

        if not self._deferred:
            background.executeLater(updater, background.Priority.BACKGROUND)
        self._deferred = True

    def visualize_additional_property(self, proxy):
        """Attempt to add an additional `PropertyProxy` to the controller. This
        should fail gracefully if the subclass has not implemented `add_proxy`.
        """
        # Check the binding type _if possible_. It's also OK when building
        # scenes to supply proxies which do not yet have a binding.
        binding = proxy.binding
        if binding is not None and not isinstance(binding, self._binding_type):
            return False  # Disallow unsupported binding types

        if proxy in self.proxies:
            return False  # Disallow duplicates

        try:
            if not self.add_proxy(proxy):
                return False
            # Only if `add_proxy` doesn't raise an exception
            self._additional_proxies.append(proxy)
            # Append the key to data model if it's not in there yet
            if proxy.key not in self.model.keys:
                self.model.keys.append(proxy.key)
            if self._showing:
                proxy.start_monitoring()
            self._proxy_update(proxy)
            return True  # The only successful exit from this method!
        except NotImplementedError:
            # Forget about it!
            return False

    def remove_additional_property(self, proxy):
        """Remove an additional property proxy"""
        if proxy not in self.proxies:
            return False  # Not there

        try:
            if not self.remove_proxy(proxy):
                return False
            if self._showing:
                # Stop monitoring if necessary
                proxy.stop_monitoring()

            # And finally stop tracking of this proxy
            self._additional_proxies.remove(proxy)
            self.model.keys.remove(proxy.key)
            return True
        except NotImplementedError:
            # Forget about it!
            return False

    def _proxy_update(self, proxy):
        if proxy.binding is None:
            return

        has_data = (proxy.binding.timestamp is not None)
        if has_data or get_class_const_trait(type(self), '_can_show_nothing'):
            self.value_update(proxy)

        state_binding = proxy.root_proxy.state_binding
        has_state = (state_binding is not None
                     and get_binding_value(state_binding))
        if has_state:
            self.state_update(proxy)

    def on_decline(self):
        """Implemented by subclasses to receive notifications that the value
        was declined.
        """

    def clear_widget(self):
        """Implemented by subclasses to receive notifications that the root
        proxy went offline."""

    def setEnabled(self, enable):
        """Implemented by subclasses to update controller as requested by
        a change in the access level.
        """
        if get_class_const_trait(type(self), '_can_edit'):
            self.widget.setEnabled(enable)

    def getInstanceId(self):
        """Retrieve the `instanceId` of the root proxy of this controller

        Note: Function added with Karabo 2.16.X
        """
        return self.proxy.root_proxy.device_id

    # -------------------------------------------------------------------------
    # Traits

    def _get_proxies(self):
        return (self.proxy,) + tuple(self._additional_proxies)

    @on_trait_change('proxy:binding,_additional_proxies:binding',
                     post_init=True)
    def _proxy_binding_update(self, proxy, name, binding):
        # One of the attached proxies got a new binding which is not None
        if binding is not None:
            self.binding_update(proxy)

    @on_trait_change('proxies.binding.config_update')
    def _proxy_value_update(self, binding, name, event):
        if self.widget is None or name != 'config_update':
            return

        try:
            # One of the attached proxies got a new value on its binding
            proxy = [p for p in self.proxies if p.binding is binding][0]
            if get_binding_value(proxy) is not None:
                self.value_update(proxy)
        except IndexError:
            pass

    @on_trait_change('proxies.root_proxy.state_binding.config_update')
    def _proxy_state_update(self, binding, name, event):
        if self.widget is None or name != 'config_update':
            return

        try:
            # One of the attached proxies got a new state on its device
            proxy = [p for p in self.proxies
                     if p.root_proxy.state_binding is binding][0]
            # Since this notification was from a different binding, we have
            # to check that the proxy's binding is there.
            if proxy.binding:
                self.state_update(proxy)
        except IndexError:
            pass

    def _widget_changed(self, new):
        """Tell the controller about its proxies whenever it gets a new widget.
        """
        if new is None:
            return

        for proxy in self.proxies:
            if proxy.binding is None:
                continue
            self.binding_update(proxy)



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on February 27, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from traits.api import Bool, Event, HasStrictTraits, Instance, List, Property

from karabo.common.api import walk_traits_object
from karabo.common.project.api import BaseProjectObjectModel, ProjectModel


class ProjectLoadingWatcher(HasStrictTraits):
    """An object which can be set as a traits notification handler for projects
    which are loading asynchronously from the database.
    """
    # A project to be loaded
    project = Instance(ProjectModel)
    # A property which can be checked by external code
    finished = Property(Bool)
    # When this event fires, the project is loaded!
    finished_event = Event

    # Project objects which are not yet loaded
    _loading_objects = List(Instance(BaseProjectObjectModel))

    def _get_finished(self):
        return len(self._loading_objects) == 0

    def _project_changed(self):
        if self.project is not None:
            self._watch_loading_object(self.project)

    def _notification_handler(self, obj, name, old, new):
        """Traits notification handler for loading objects.
        """
        if new:
            self._loading_objects.remove(obj)
            obj.on_trait_change(self._notification_handler, 'initialized',
                                remove=True)

            # It's possible that this object introduced more children which
            # now need to be watched
            self._watch_loading_object(obj)

        # Are we done yet?
        if self.finished:
            self.finished_event = True

    def _watch_loading_object(self, project_object):
        """Add a whole bunch of traits listeners to a project object and its
        children so that it can be determined when all of its children are
        loaded.
        """
        def visitor(model):
            if isinstance(model, BaseProjectObjectModel):
                if not model.initialized:
                    model.on_trait_change(self._notification_handler,
                                          'initialized')
                    self._loading_objects.append(model)

        walk_traits_object(project_object, visitor)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import BarGraphItem, ScatterPlotItem

from karabogui.graph.common.api import make_brush, safe_log10


class VectorBarGraphPlot(BarGraphItem):
    """VectorBarGraphPlot Item Wrapper for a standard interface
    """

    def __init__(self, width, brush):
        X_DEFAULT, Y_DEFAULT = [], []
        super().__init__(
            x=X_DEFAULT, width=width, height=Y_DEFAULT, y0=0, brush=brush)

    def setOpts(self, **opts):
        if "logMode" not in self.opts:
            self.opts["logMode"] = [False, False]
        super().setOpts(**opts)

    def getData(self):
        """Returns the corrected x and y data from the bar attributes."""
        # Calculate x-data from x
        x_data = np.array(self.opts['x'])
        if self.opts['logMode'][0]:
            x_data = safe_log10(x_data)

        # Calculate y-data from y0 and height
        height, y0 = self.opts['height'], self.opts['y0']
        y_data = np.array(height) - y0
        if self.opts['logMode'][1]:
            y_data = safe_log10(y_data)

        return x_data, y_data

    def setData(self, x, y):
        self.setOpts(x=x, height=y, y0=0)

    def set_width(self, value):
        self.setOpts(width=value)

    # -----------------------------------------------------------------------
    # Qt methods

    def _getNormalizedCoords(self):
        def asarray(x):
            if x is None or np.isscalar(x) or isinstance(x, np.ndarray):
                return x
            return np.array(x)

        # !----- Start modification -----!
        # MOD: Utilize corrected x- and y-data for the bar geometries
        x_data, y_data = self.getData()

        # MOD: Utilize x-data for the x0 calculation
        x = asarray(x_data)
        x0 = asarray(self.opts.get('x0'))
        width = asarray(self.opts.get('width'))

        # MOD: Calculate x0 based on x-data
        if x0 is None:
            if width is None:
                raise Exception('must specify either x0 or width')
            # MOD: Remove x1 calculation
            if x is not None:
                x0 = x - width / 2.
            else:
                # MOD: Remove x1 requirement on exception message
                raise Exception('must specify either x0 or width')
        # MOD: Remove width calculation

        # MOD: Remove y and y1 requirements
        y0 = asarray(self.opts.get('y0'))
        # MOD: Use the corrected y-data instead of height
        height = asarray(y_data)

        # MOD: Use a relative starting point and values if log-y is enabled.
        # We consider values < 0 as nans
        if self.opts['logMode'][1] and height.size:
            order = np.floor(np.nanmin(height))
            y0 = order - 1
            height -= y0
            # Set height of invalid values as 0
            height[~np.isfinite(height)] = 0

        # !----- End modification -----!
        # ensure x0 < x1 and y0 < y1
        t0, t1 = x0, x0 + width
        x0 = np.minimum(t0, t1, dtype=np.float64)
        x1 = np.maximum(t0, t1, dtype=np.float64)
        t0, t1 = y0, y0 + height
        y0 = np.minimum(t0, t1, dtype=np.float64)
        y1 = np.maximum(t0, t1, dtype=np.float64)

        # here, all of x0, y0, x1, y1 are numpy objects,
        # BUT could possibly be numpy scalars
        return x0, y0, x1, y1

    def setLogMode(self, xMode, yMode):
        """Stores the log mode. This is called by the viewbox."""
        log_mode = [xMode, yMode]
        if self.opts.get('logMode', None) == log_mode:
            return
        self.setOpts(logMode=log_mode)


class ScatterGraphPlot(ScatterPlotItem):
    """ScatterGraphPlot Item to display the points in a Cloud

    NOTE: The most recent value has a ``firebrick`` brush!
    """
    last_value_brush = make_brush('r', 200)
    points_brush = make_brush('b', 200)

    def __init__(self, pen, cycle, **kwargs):
        self._cycle = cycle
        default_x, default_y = [], []
        super().__init__(
            x=default_x, y=default_y, pxMode=True, pen=pen, **kwargs)

    def setData(self, x, y, *args, **kwargs):
        """Set the data of the scatter plot"""
        if len(x) != len(y):
            return

        if self._cycle:
            size = len(x)
            if not size:
                super().setData(
                    x, y, brush=self.points_brush, **kwargs)
                return

            brush = []
            if size > 1:
                brush.extend([self.points_brush] * (size - 1))
            brush.append(self.last_value_brush)
        else:
            brush = self.points_brush

        super().setData(x, y, brush=brush, **kwargs)



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on July 21, 2013
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QEvent, Qt, QTimer, Slot
from qtpy.QtWidgets import (
    QPlainTextEdit, QPushButton, QTextEdit, QVBoxLayout, QWidget)

from karabo.native import Hash, create_html_hash
from karabogui.util import create_table_string

_PADDING = 5


class PopupWidget(QWidget):
    def __init__(self, can_freeze: bool = False, parent: QWidget = None):
        super().__init__(parent=parent, flags=Qt.Dialog)
        flags = Qt.WindowCloseButtonHint | Qt.WindowStaysOnTopHint
        self.setWindowFlags(self.windowFlags() | flags)
        self.setWindowModality(Qt.NonModal)
        self._ui_info = TextEdit(self)
        self._ui_info.setReadOnly(True)

        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(1, 1, 1, 1)
        main_layout.addWidget(self._ui_info)

        # default is updating the pop-up
        self.can_freeze = can_freeze
        self.freeze = False
        self.info_cache = None

        # NOTE: The pop-up is used by the configurator and the navigation
        # Configurator requires freezing!
        if can_freeze:
            self._ui_freeze_button = QPushButton()
            self._ui_freeze_button.setFocusPolicy(Qt.NoFocus)
            self._ui_freeze_button.setToolTip("Freeze the pop-up information")
            self._ui_freeze_button.setStyleSheet("text-align: center")
            self._ui_freeze_button.clicked.connect(self.toggle_freeze)
            main_layout.addWidget(self._ui_freeze_button)

            # update our status!
            self.update_button_status()

        self.setWindowTitle(" ")

    def update_button_status(self):
        button_text = "Update" if self.freeze else "Freeze"
        self._ui_freeze_button.setText(button_text)

    @property
    def text(self):
        return self._ui_info.toPlainText()

    @Slot()
    def toggle_freeze(self):
        self.freeze = not self.freeze
        if self.freeze is False and self.info_cache is not None:
            self.setInfo(self.info_cache)
        self.update_button_status()

    def setInfo(self, info: dict | Hash):
        if self.freeze:
            # We cache the lastet update for later display when the
            # freeze is toggled!
            self.info_cache = info
            return

        if isinstance(info, Hash):
            html_string = create_html_hash(info)
            height = len(info.paths())
        else:
            height = len(info)
            html_string = create_table_string(info)

        self._ui_info.setHtml(html_string)
        self._ui_info.fitHeightToContent(height + _PADDING)
        # Restore scrolling position to prevent irritating scrolling up while
        # updating the popup dialog with further information on an expected
        # parameter
        scrollBar = self._ui_info.verticalScrollBar()
        pos = scrollBar.sliderPosition()
        scrollBar.setValue(pos)

    def reset(self):
        self.freeze = False
        self.info_cache = None
        self.update_button_status()

    def closeEvent(self, event: QEvent):
        """Reimplemented function from Qt

           Erase all caching information and set the freeze to false
        """
        if self.can_freeze:
            self.reset()
        super().closeEvent(event)


class TextEdit(QTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._fittedWidth = 310
        self._fittedHeight = 0

    def sizeHint(self):
        sizeHint = QTextEdit.sizeHint(self)
        sizeHint.setWidth(self._fittedWidth)
        sizeHint.setHeight(self._fittedHeight)
        return sizeHint

    def fitHeightToContent(self, nbInfoKeys: int):
        self._fittedHeight = self.fontMetrics().height() * nbInfoKeys
        self.updateGeometry()


class TextPopupWidget(QWidget):
    """A popup widget to show the text like tooltip."""
    def __init__(self, model, parent=None):
        super().__init__(parent=parent)
        self.setWindowFlags(self.windowFlags() | Qt.Popup)

        self.setGeometry(self.x(), self.y(),
                         model.popup_width, model.popup_height)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.text_edit = QPlainTextEdit(parent=self)
        self.text_edit.setReadOnly(True)
        if model.text:
            self.text_edit.setPlainText(model.text)
        else:
            self.text_edit.setPlaceholderText(
                "Add text by double-clicking on the widget or by 'Edit Text' "
                "right click menu.")
        layout.addWidget(self.text_edit)

        self._ticker = QTimer(self)
        self._ticker.setInterval(10000)  # 10s
        self._ticker.setSingleShot(True)
        self._ticker.timeout.connect(self.close)
        self._ticker.start()

        self.setStyleSheet("""
            QPlainTextEdit {
                background-color: rgba(255, 255, 226, 200);
                border: 1px solid black;}""")

    def closeEvent(self, event):
        """Close the ticker before closing the popup widget."""
        if self._ticker.isActive():
            self._ticker.stop()
        super().closeEvent(event)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import importlib
import re
import sys
from argparse import ArgumentParser
from collections import defaultdict
from subprocess import STDOUT, CalledProcessError, check_output

import requests
from packaging.version import Version
from qtpy import uic
from qtpy.QtCore import (
    QAbstractAnimation, QEasingCurve, QProcess, QPropertyAnimation, Qt, Slot)
from qtpy.QtWidgets import (
    QDialog, QHeaderView, QProgressBar, QPushButton, QTableWidgetItem)

from karabogui import icons
from karabogui.controllers.util import load_extensions
from karabogui.dialogs.utils import get_dialog_ui
from karabogui.singletons.api import get_config

_TIMEOUT = 0.5
_TAG_REGEX = r"^\d+\.\d+\.\d+(\.\d+)?$"
_PKG_NAME = "GUIExtensions"
_PKG_NAME_ALL = "GUIExtensions[all]"
_PACKAGE_URL = "https://git.xfel.eu/api/v4/projects/5187/packages"
_PYPI_INDEX = f"{_PACKAGE_URL}/pypi/simple"
UNDEFINED_VERSION = "Undefined"

DURATION = 500  # [ms]


class LoadingBar(QProgressBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTextVisible(False)
        self.animation = QPropertyAnimation(
            targetObject=self, propertyName=b"value")
        self.animation.setStartValue(self.minimum())
        self.animation.setEndValue(self.maximum())
        self.animation.finished.connect(self._update_bar)
        self.animation.setDuration(DURATION)
        self.animation.setEasingCurve(QEasingCurve.InOutQuad)
        self.animation.start()

    @Slot()
    def _update_bar(self):
        if self.animation.currentValue() == self.maximum():
            self.animation.setDirection(QAbstractAnimation.Backward)
            self.setInvertedAppearance(True)
            self.animation.start()
        elif self.animation.currentValue() == self.minimum():
            self.animation.setDirection(QAbstractAnimation.Forward)
            self.setInvertedAppearance(False)
            self.animation.start()


def create_item(name: str) -> QTableWidgetItem:
    """Create a read only standard table widget item"""
    item = QTableWidgetItem(name)
    flags = item.flags()
    flags &= ~ Qt.ItemIsEditable
    item.setFlags(flags)
    return item


def get_pkg_version(name: str) -> str:
    """Gets the current version of the package"""
    try:
        return importlib.metadata.version(name)
    except Exception:
        return UNDEFINED_VERSION


def get_index_list() -> dict:
    try:
        result = requests.get(_PACKAGE_URL, timeout=_TIMEOUT)
    except requests.Timeout:
        return {}
    package_list = result.json()

    tag_regex = re.compile(_TAG_REGEX)
    uniques = defaultdict(list)
    for p in package_list:
        name = p["name"]
        version = p["version"]
        if not tag_regex.match(version):
            continue
        uniques[name].append(version)

    def safe_version(semver: str):
        try:
            return Version(semver)
        except Exception:
            return Version('0.0.0')

    return {name: max(versions, key=safe_version)
            for name, versions in uniques.items()}


def uninstall_package(package: str) -> str:
    try:
        output = check_output([sys.executable, "-m", "pip",
                               "uninstall", "--yes", package],
                              stderr=STDOUT)
        return output.decode()
    except CalledProcessError as e:
        error_message = e.output.decode("utf-8")
        return (f"Error uninstalling package. Is it installed?\n"
                f"{error_message}")


def install_package(package: str, version: str) -> str:
    """Installs a given wheel file"""
    try:
        output = check_output(
            [sys.executable, "-m", "pip", "install", f"{package}=={version}",
             "--force-reinstall", "--disable-pip-version-check",
             "--index-url", _PYPI_INDEX],
            stderr=STDOUT)
        # Reload the entry points
        load_extensions()
        return output.decode()
    except CalledProcessError as e:
        error_message = e.output.decode("utf-8")
        return f"Error installing {_PKG_NAME} package:\n{error_message}"


def install_specific_tag(package: str, tag: str) -> str:
    """Updates the package to the given tag version.

    :returns str:
        The output given by the process.
    """
    assert isinstance(tag, str)
    # Install it using a system call
    return install_package(package, tag)


def is_package_updated(package_name: str) -> bool:
    """Checks if there are new extensions available"""

    latest_version = get_index_list().get(package_name, "0.0.0")
    current_version = get_pkg_version(package_name)
    if current_version == UNDEFINED_VERSION:
        return True
    return Version(latest_version) > Version(current_version)


class UpdateDialog(QDialog):
    """Update dialog for the Karabo GUI extensions package

    This dialog is used for checking the current and latest versions of
    karabo-extensions package. The user may also update the package, where a
    pip install install call is done.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("update_dialog.ui"), self)
        self.setModal(False)
        self.loadingbar.setVisible(False)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.label_current.setText(UNDEFINED_VERSION)
        self.label_latest.setText(UNDEFINED_VERSION)
        self.button_refresh.setIcon(icons.refresh)

        self.button_update.setEnabled(False)
        self.button_close.clicked.connect(self.accept)

        self._package_index = None
        # Store the current running process
        self._process = None
        self.extension_url.setText(
            "The extensions can be installed via PyPi Index: "
            f"<b>{_PYPI_INDEX}</b>")
        self.refresh_versions()
        self.look_for_update.setChecked(get_config()["check_updates"])
        self.look_for_update.clicked.connect(self.check_update_on_startup)

    @property
    def package_index(self):
        """Cache the package index to only load once"""
        if self._package_index is None:
            self._package_index = get_index_list()
        return self._package_index

    def _fill_package_overview(self):
        self.tableWidget.clearContents()
        self.tableWidget.setRowCount(0)

        packages = get_index_list()
        labels = ["Name", "Current Version", "Available Version", "Install",
                  "Uninstall"]
        self.tableWidget.setColumnCount(len(labels))
        self.tableWidget.setSelectionBehavior(self.tableWidget.SelectRows)
        self.tableWidget.setHorizontalHeaderLabels(labels)

        header = self.tableWidget.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.ResizeToContents)
        row = 0
        for name, version in packages.items():
            if name == _PKG_NAME:
                continue
            self.tableWidget.insertRow(row)
            self.tableWidget.setItem(row, 0, create_item(name))
            current_version = get_pkg_version(name)
            self.tableWidget.setItem(row, 1, create_item(current_version))
            self.tableWidget.setItem(row, 2, create_item(version))

            install = QPushButton(icons.mediaStart, "Install")
            install.clicked.connect(self._table_button_clicked)
            info = {"package": name,
                    "mode": "install",
                    "version": version}
            install.setProperty("items", info)
            install.setEnabled(version != current_version)
            self.tableWidget.setCellWidget(row, 3, install)

            uninstall = QPushButton(icons.mediaBackward, "Uninstall")
            uninstall.setEnabled(current_version != UNDEFINED_VERSION)
            info = {"package": name,
                    "mode": "uninstall"}
            uninstall.setProperty("items", info)
            uninstall.clicked.connect(self._table_button_clicked)
            self.tableWidget.setCellWidget(row, 4, uninstall)
            row += 1

    @Slot()
    def _table_button_clicked(self):
        """An install or uninstall button was clicked in the tablewidget"""
        sender = self.sender()
        info = sender.property("items")
        mode = info["mode"]
        name = info["package"]
        self.tableWidget.setEnabled(False)
        if mode == "install":
            self._set_buttons_install()
            self._start_package_install(name, info["version"])
        if mode == "uninstall":
            self._set_buttons_uninstall()
            self._start_package_uninstall(name)

    def refresh_versions(self):
        """Refreshing the current and latest package versions"""
        self._update_current_version()
        self._update_latest_version()
        current = self.label_current.text()
        latest = self.label_latest.text()
        needs_updating = current != latest != UNDEFINED_VERSION
        self.button_update.setEnabled(needs_updating)
        self.button_uninstall.setEnabled(current != UNDEFINED_VERSION)
        self._fill_package_overview()

    def _update_current_version(self):
        """Updates the current version of the device"""
        self.label_current.setVisible(False)

        current_version = get_pkg_version(_PKG_NAME)
        self._update_log("Version refreshed")

        self.label_current.setVisible(True)
        self.label_current.setText(current_version)

    def _update_latest_version(self):
        """Updates the latest version of the device"""
        self.label_latest.setVisible(False)
        version = self.package_index.get(_PKG_NAME, UNDEFINED_VERSION)
        self.label_latest.setVisible(True)
        self.label_latest.setText(version)

    def _start_process(self, cmd):
        """Starts a process with the given arguments"""
        if self._process is not None:
            self._clear_process()

        if self._process is None:
            # Create a process and connect its
            self._process = QProcess(self)
            self._process.setProcessChannelMode(QProcess.MergedChannels)
            self._process.readyRead.connect(self._on_output)
            self._process.finished.connect(self._on_finished)
            self.loadingbar.setVisible(True)
            self._process.start(cmd)

    def _start_package_install(self, package, tag):
        """Create a QProcess to update to the latest tag.

        This process' signals are connected to the given callbacks."""
        cmd = (f"pip install {package}=={tag} "
               f"--force-reinstall --index-url {_PYPI_INDEX} "
               "--disable-pip-version-check")
        self._start_process(cmd)

    def _start_package_uninstall(self, package):
        """Uninstalls the current GUIExtensions package"""
        cmd = f"pip uninstall --yes {package}"
        self._start_process(cmd)

    def _update_log(self, text):
        self.text_log.append(f"> {text}")

    def _clear_process(self):
        """Clears all running process resources"""
        if self._process is not None:
            self._process.kill()

        self._process = None

    def done(self, result):
        self._clear_process()
        return super().done(result)

    # --------------------------------------------------------------------
    # Qt Slots

    @Slot()
    def on_button_refresh_clicked(self):
        self.button_refresh.setEnabled(False)
        self.refresh_versions()
        self.button_refresh.setEnabled(True)

    @Slot()
    def on_button_uninstall_clicked(self):
        self._set_buttons_uninstall()
        self._start_package_uninstall(_PKG_NAME)

    @Slot()
    def on_button_update_clicked(self):
        """Updates gui extensions to the latest tag"""
        self._set_buttons_install()
        tag = self.label_latest.text()
        self._start_package_install(_PKG_NAME, tag)

    @Slot()
    def _on_output(self):
        while self._process.canReadLine():
            line = self._process.readLine()
            data = line.data().decode("utf-8").strip()
            self._update_log(data)

    @Slot()
    def _on_finished(self):
        """Called when the uninstall finishes or crashes"""
        self._clear_process()
        self.refresh_versions()
        load_extensions()
        self._set_buttons_finish()
        self.loadingbar.setVisible(False)

    # Buttons
    # -----------------------------------------------------------------------

    def _set_buttons_uninstall(self):
        self.button_uninstall.setEnabled(False)
        self.button_update.setEnabled(False)
        self.tableWidget.setEnabled(False)

    def _set_buttons_install(self):
        self.tableWidget.setEnabled(False)
        self.button_refresh.setEnabled(False)
        self.button_update.setEnabled(False)
        self.button_uninstall.setEnabled(False)

    def _set_buttons_finish(self):
        self.tableWidget.setEnabled(True)
        self.button_refresh.setEnabled(True)

    @Slot(bool)
    def check_update_on_startup(self, toggled):
        """Check if the user wants to check for updates on startup."""
        get_config()['check_updates'] = toggled


def main():
    description = """Command-line tool to update the {} package"""
    ap = ArgumentParser(description=description.format(_PKG_NAME))
    ap.add_argument("-t", "--tag", nargs=1, type=str,
                    help="Desired package version (tag)")
    ap.add_argument("-u", "--uninstall", action="store_true", required=False,
                    help="Uninstalls any previous versions")
    ap.add_argument("-l", "--latest", action="store_true", required=False,
                    help="Upgrade to the latest tag")
    ap.add_argument("-V", "--version", action="store_true", required=False,
                    help="Print the installed version")
    ap.add_argument("-a", "--all", action="store_true", required=False,
                    help="Chose if all dependencies should be installed")
    args = ap.parse_args()

    if not any([args.uninstall, args.tag, args.latest, args.version]):
        print("At least one option must be given! Please use -h for help.")
        return

    if args.version:
        version = get_pkg_version(_PKG_NAME)
        if version == UNDEFINED_VERSION:
            print(f"{_PKG_NAME} package not installed")
        else:
            print(f"{_PKG_NAME}: version {version} installed")

    elif args.uninstall:
        print(f"Uninstalling {_PKG_NAME} package")
        output = uninstall_package(_PKG_NAME)
        print(output)

    elif args.tag:
        tag = args.tag[0]
        package = _PKG_NAME if not args.all else _PKG_NAME_ALL
        print(f"Installing {package} version {tag}\n")
        output = install_specific_tag(package, tag)
        print(output)

    elif args.latest:
        package_list = get_index_list()
        tag = package_list.get(_PKG_NAME, UNDEFINED_VERSION)
        package = _PKG_NAME if not args.all else _PKG_NAME_ALL
        print(f"Installing {package} version {tag}\n")
        output = install_specific_tag(package, tag)
        print(output)


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from datetime import datetime

from pyqtgraph import InfiniteLine, SignalProxy
from qtpy.QtCore import QObject, Slot

from karabogui import icons
from karabogui.graph.common.api import (
    AlarmAxisItem, CoordsLegend, StateAxisItem, TimeAxisItem, create_button,
    float_to_string, get_alarm_string, get_state_string, make_pen)


class CrossTargetController(QObject):
    def __init__(self, plotItem):
        """Controller for showing a cross target on a plot"""
        super().__init__()
        self.plotItem = plotItem

        self.v_line = None
        self.h_line = None
        self.proxy = None
        self.legend = None
        self.action_button = create_button(
            checkable=True,
            icon=icons.target,
            tooltip="Get a CrossTarget for the plot",
            on_clicked=self.toggle)

    @Slot(object)
    def mouseMoved(self, event):
        """Catch the mouseMove event on the plotItem and show coordinates"""
        pos = event[0]
        # using signal proxy turns original arguments into a tuple
        if self.plotItem.sceneBoundingRect().contains(pos):
            mousePoint = self.plotItem.vb.mapSceneToView(pos)
            x, y = mousePoint.x(), mousePoint.y()
            self.legend.set_value(self._get_x_value(x), self._get_y_value(y))
            self.v_line.setPos(x)
            self.h_line.setPos(y)
            self.legend.setVisible(True)
        else:
            self.legend.setVisible(False)

    def _get_x_value(self, value):
        """Get the real x value depending on the axis"""
        plotItem = self.plotItem
        axis = plotItem.getAxis("bottom")
        if isinstance(axis, TimeAxisItem):
            x_min, x_max = axis.range
            difference = x_max - x_min
            if difference < 60:
                fmt = "%H:%M:%S,%f"
            elif difference < 3600 * 24:
                fmt = "%H:%M:%S"
            elif difference < 3600 * 24 * 30:
                fmt = "%d/%m, %H:%M:%S"
            elif difference < 3600 * 24 * 30 * 365:
                fmt = "%d/%m, %H:%M"
            else:
                fmt = "%d/%m/%Y"
            timestamp = datetime.fromtimestamp(value)
            return timestamp.strftime(fmt)
        elif axis.logMode:
            return convert_log(value)
        return float_to_string(value, precision=3)

    def _get_y_value(self, value):
        """Get the real y value depending on the axis"""
        plotItem = self.plotItem
        axis = plotItem.getAxis("left")
        if isinstance(axis, StateAxisItem):
            return get_state_string(value)
        elif isinstance(axis, AlarmAxisItem):
            return get_alarm_string(value)
        elif axis.logMode:
            return convert_log(value)
        return float_to_string(value, precision=3)

    # -----------------------------------------------------------------------
    # Public methods

    @Slot(bool)
    def toggle(self, state):
        if state:
            self.activate()
        else:
            self.deactivate()

    def activate(self):
        pen = make_pen("a")
        self.v_line = InfiniteLine(angle=90, pen=pen, movable=False)
        self.h_line = InfiniteLine(angle=0, pen=pen, movable=False)
        self.plotItem.addItem(self.v_line, ignoreBounds=True)
        self.plotItem.addItem(self.h_line, ignoreBounds=True)
        self.legend = CoordsLegend(color='k')
        self.legend.setParentItem(self.plotItem.vb)
        self.legend.anchor(itemPos=(1, 0), parentPos=(1, 0), offset=(-5, 5))
        self.proxy = SignalProxy(self.plotItem.scene().sigMouseMoved,
                                 rateLimit=50, slot=self.mouseMoved)

    def deactivate(self):
        self.proxy.disconnect()
        self.proxy = None
        for item in [self.h_line, self.v_line]:
            self.plotItem.removeItem(item)
            item.deleteLater()
        self.v_line = None
        self.h_line = None
        self.legend.setParentItem(None)
        self.legend.deleteLater()
        self.legend = None


def convert_log(value):
    """Return the e-annotation of the given value with 3 precisions."""
    try:
        return f"{10 ** value:.3e}"
    except Exception:
        return str(value)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from abc import abstractmethod
from io import StringIO
from operator import sub

from qtpy.QtCore import QMimeData, QPoint
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QApplication, QDialog, QMessageBox

from karabo.common.scenemodel.api import (
    BaseLayoutModel, BaseWidgetObjectData, SceneModel, read_scene, write_scene)
from karabogui.dialogs.api import ReplaceDialog
from karabogui.sceneview.bases import BaseSceneAction
from karabogui.sceneview.utils import add_offset, calc_relative_pos
from karabogui.util import move_to_cursor

MIME_TYPE = 'image/svg+xml'
GRID_INC = 10


def _add_models_to_clipboard(models, rect):
    """Given a list of models, add their SVG representation to the clipboard
    """
    scene = SceneModel(children=models, width=rect.width(),
                       height=rect.height())
    xml = write_scene(scene)
    mime_data = QMimeData()
    mime_data.setData(MIME_TYPE, xml.encode('utf-8'))
    QApplication.clipboard().setMimeData(mime_data)


def _read_models_from_clipboard():
    """Read SVG data from the clipboard into a SceneModel and return the child
    models.
    """
    mime_data = QApplication.clipboard().mimeData()
    if not mime_data.hasFormat(MIME_TYPE):
        return []

    byte_array = mime_data.data(MIME_TYPE)
    file_obj = StringIO(byte_array.data().decode('utf-8'))
    scene = read_scene(file_obj)
    scene.assure_svg_data()

    return scene.children


class SceneCopyAction(BaseSceneAction):
    """Copy all selected objects from the scene view
    """
    def perform(self, scene_view):
        selection_model = scene_view.selection_model
        if len(selection_model) == 0:
            return
        models = [o.model for o in selection_model]
        rect = selection_model.get_selection_bounds()
        _add_models_to_clipboard(models, rect)


class SceneCutAction(BaseSceneAction):
    """Remove all selected objects from the scene view.
    """
    def perform(self, scene_view):
        selection_model = scene_view.selection_model
        if len(selection_model) == 0:
            return

        models = [o.model for o in selection_model]
        rect = selection_model.get_selection_bounds()
        selection_model.clear_selection()

        # Make sure they can be pasted
        _add_models_to_clipboard(models, rect)

        for model in models:
            scene_view.remove_model(model)
        scene_view.update()


class BaseScenePasteAction(BaseSceneAction):
    """Base class for paste actions.

    This contains the boilerplate that all layout actions have in common.
    """
    def perform(self, scene_view):
        pasted_models = _read_models_from_clipboard()
        if len(pasted_models) == 0:
            return

        self.run_action(pasted_models, scene_view)
        scene_view.update()

    @abstractmethod
    def run_action(self, models, scene_view):
        """Implemented by derived classes to implement what should happen with
        the copied ``models``.
        """

    def add_to_scene(self, models, scene_view):
        # Map cursor to widget
        pos = scene_view.mapFromGlobal(QCursor.pos())
        rect = scene_view.visibleRegion().boundingRect()
        if rect.contains(pos):
            new_origin = (int(pos.x()), int(pos.y()))
        else:
            # Get the relative
            new_origin = (rect.x() + 10, rect.y() + 10)
        old_origin = calc_relative_pos(models)
        x_off, y_off = map(sub, new_origin, old_origin)

        # Add offset on models
        for model in models:
            add_offset(model, x=x_off, y=y_off)

        scene_view.add_models(*models)
        scene_view.select_models(*models)


class ScenePasteAction(BaseScenePasteAction):
    """Paste the contents of the clipboard to the scene view.
    """
    def run_action(self, models, scene_view):
        """Add ``models`` to scene view."""
        # Add some offset on the pasted model to help users see that there are
        # new objects from the command
        if not len(models):
            # Do nothing
            return

        self.add_to_scene(models, scene_view)


class ScenePasteReplaceAction(BaseScenePasteAction):
    """Paste and replace the contents of the clipboard to the scene view.
    """
    def _get_widget_model_keys(self, model, keys_list):
        """This recursive method fills the given ``key_list`` with model keys
        of ``BaseWidgetObjectData`` types.
        """
        if isinstance(model, BaseLayoutModel):
            for child in model.children:
                self._get_widget_model_keys(child, keys_list)
        elif isinstance(model, BaseWidgetObjectData):
            keys_list.extend(model.keys)

    def _set_widget_model_keys(self, model, mapped_device_ids):
        if isinstance(model, BaseLayoutModel):
            for child in model.children:
                self._set_widget_model_keys(child, mapped_device_ids)
        elif isinstance(model, BaseWidgetObjectData):
            new_key_list = []
            for key in model.keys:
                device_id, property_key = key.split('.', 1)
                new_device_id = mapped_device_ids.get(device_id)
                if new_device_id is None:
                    new_key = key
                else:
                    new_key = f"{new_device_id}.{property_key}"
                new_key_list.append(new_key)
            # Overwrite key list with new keys
            model.keys = new_key_list

    def run_action(self, models, scene_view):
        """Modify the device IDs of ``models``and add them to the scene view.
        """
        if not len(models):
            # Do nothing
            return

        keys = []
        for m in models:
            self._get_widget_model_keys(m, keys)

        device_ids = sorted({k.split('.', 1)[0] for k in keys})
        dialog = ReplaceDialog(device_ids, parent=scene_view)
        if dialog.exec() != QDialog.Accepted:
            return

        mapped_device_ids = dialog.mappedDevices()
        for m in models:
            self._set_widget_model_keys(m, mapped_device_ids)
        self.add_to_scene(models, scene_view)


class SceneSelectAllAction(BaseSceneAction):
    """Select all objects in the scene view.
    """
    def perform(self, scene_view):
        scene_view.select_all()


class SceneDeleteAction(BaseSceneAction):
    """Delete selected objects from the scene view.
    """
    def perform(self, scene_view):
        selection_model = scene_view.selection_model
        if len(selection_model) == 0:
            return

        dialog = QMessageBox(
            QMessageBox.Question, "Really delete?", "Do you really "
            "want to delete the items?", QMessageBox.Yes | QMessageBox.No,
            parent=scene_view)
        move_to_cursor(dialog)

        if dialog.exec() == QMessageBox.No:
            return

        for o in selection_model:
            scene_view.remove_model(o.model)
        selection_model.clear_selection()


class SceneMoveAction(BaseSceneAction):
    """Move selected objects inside the scene view.
    """
    def perform(self, scene_view):
        selection_model = scene_view.selection_model
        if len(selection_model) == 0:
            return

        inc = GRID_INC if scene_view.snap_to_grid else 1
        increment = {"Left": QPoint(-inc, 0),
                     "Right": QPoint(inc, 0),
                     "Up": QPoint(0, -inc),
                     "Down": QPoint(0, inc)}
        offset = increment.get(self.text)
        # move each item in the selection with increment!
        for c in scene_view.selection_model:
            c.translate(offset)
        scene_view.update()


class SceneAlignAction(BaseSceneAction):
    """Align selected objects inside the scene view.
    """
    def perform(self, scene_view):
        selection_model = scene_view.selection_model
        if len(selection_model) == 0:
            return

        align_coord = {
            "Left": min([c.model.x for c in selection_model]),
            "Right": max(
                [c.model.x + c.model.width for c in selection_model]),
            "Top": min([c.model.y for c in selection_model]),
            "Bottom": max(
                [c.model.y + c.model.height for c in selection_model])

        }
        for c in selection_model:
            if self.text == "Left":
                offset = QPoint(align_coord["Left"] - c.model.x, 0)
            elif self.text == "Right":
                offset = QPoint(
                    align_coord["Right"] - c.model.x - c.model.width, 0)
            elif self.text == "Top":
                offset = QPoint(0, align_coord["Top"] - c.model.y)
            elif self.text == "Bottom":
                offset = QPoint(
                    0, align_coord["Bottom"] - c.model.y - c.model.height)
            c.translate(offset)

        scene_view.update()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os
import sys

from qtpy.QtCore import QLocale

from karabogui.const import IS_MAC_SYSTEM
from karabogui.programs.base import create_gui_app, init_gui
from karabogui.testing import set_test_organization_info


def test_start_app(mocker):
    """Make sure that we can start a gui application"""
    os.environ["KARABO_TEST_GUI"] = "1"
    info = mocker.patch("karabogui.programs.base.set_app_info")
    app = create_gui_app(sys.argv)
    info.assert_called_with(app, "XFEL", "xfel.eu", "KaraboGUI")
    init_gui(app, use_splash=True)
    set_test_organization_info(app)
    font = app.font()
    assert font.family() == "Source Sans Pro"
    psize = 10 if not IS_MAC_SYSTEM else 13
    assert font.pointSize() == psize
    locale = QLocale()
    assert locale.country() == QLocale.UnitedStates
    assert locale.language() == QLocale.English
